// Copyright (c) 2007 Paul Miner <$firstname.$lastname@gmail.com>
// This software is provided 'as-is', without any express or implied warranty. In no event will the authors be held liable for any damages arising from the use of this software.
// Permission is granted to anyone to use this software for any purpose, including commercial applications, and to alter it and redistribute it freely, subject to the following restrictions:
//     1. The origin of this software must not be misrepresented; you must not claim that you wrote the original software. If you use this software in a product, an acknowledgment in the product documentation would be appreciated but is not required.
//     2. Altered source versions must be plainly marked as such, and must not be misrepresented as being the original software.
//     3. This notice may not be removed or altered from any source distribution.

#define UNICODE
#define _UNICODE
#define _WIN32_WINNT 0x0500
#define JNI_FALSE 0
#define JNI_TRUE 1

#include <windows.h>
#include <winioctl.h>
#include <tchar.h>
#include "org_blinkenbyte_io_WindowsBlockDevice.h"

// Given a class name and error message, throw an exception
void throwByName(JNIEnv* env, const char* className, const char* msg)
{
	jclass exceptionClass = (*env)->FindClass(env, className);
	if (exceptionClass != NULL)
	{
		(*env)->ThrowNew(env, exceptionClass, msg);
		(*env)->DeleteLocalRef(env, exceptionClass);
	}
	return;
}

// This is called if MultiByteToWideChar() returns an error, and throws an exception
void handleMultiByteToWideCharError(JNIEnv* env)
{
	DWORD errorCode = GetLastError();
	const char* errorMsg = "Unknown error in MultiByteToWideChar";
	switch (errorCode)
	{
	case ERROR_INSUFFICIENT_BUFFER:
		errorMsg = "Insufficient buffer in MultiByteToWideChar";
		break;

	case ERROR_INVALID_FLAGS:
		errorMsg = "Invalid flags in MultiByteToWideChar";
		break;

	case ERROR_INVALID_PARAMETER:
		errorMsg = "Invalid parameter passed to MultiByteToWideChar";
		break;

	case ERROR_NO_UNICODE_TRANSLATION:
		errorMsg = "No unicode translation in MultiByteToWideChar";
		break;
	}
	throwByName(env, "java/lang/RuntimeException", errorMsg);
	return;
}

// Given a Java byte array, copy it into a HANDLE and return it
HANDLE convertByteArrayToHandle(JNIEnv* env, jbyteArray handleBytes)
{
	HANDLE returnHandle = 0;
	if (handleBytes == NULL)
		return 0;
	if ((*env)->GetArrayLength(env, handleBytes) != sizeof(returnHandle))
		return 0;
	(*env)->GetByteArrayRegion(env, handleBytes, 0, sizeof(returnHandle), (BYTE*)(&returnHandle));
	return returnHandle;
}

// Wrapper for convertByteArrayToHandle() that throws an exception if the handle is invalid
// Returns 0 if exception thrown
HANDLE getHandle(JNIEnv* env, jbyteArray handleBytes)
{
	HANDLE fileHandle = convertByteArrayToHandle(env, handleBytes);
	if ((*env)->ExceptionOccurred(env) != NULL)
		return 0;
	if ((fileHandle == INVALID_HANDLE_VALUE) || (fileHandle == NULL))
	{
		throwByName(env, "java/lang/RuntimeException", "Invalid handle");
		return 0;
	}
	return fileHandle;
}

// Open the specified file, with various sync options
// If sectorMode is enabled, do not use buffering (we do not want read-ahead if possible)
// Return the file handle in a Java byte[]
JNIEXPORT jbyteArray JNICALL Java_org_blinkenbyte_io_WindowsBlockDevice_nativeOpen(JNIEnv* env, jclass cls, jstring filename, jboolean allowWrite, jboolean syncWriteData, jboolean syncWriteMetaData, jboolean sectorMode)
{
	const jbyte* utf8Filename;
	int wcFilenameLen;
	WCHAR* wcFilename;
	HANDLE fileHandle;
	jbyteArray returnHandle = NULL;

	if (filename == NULL)
	{
		throwByName(env, "java/lang/NullPointerException", "Filename is null");
		return NULL;
	}
	utf8Filename = (*env)->GetStringUTFChars(env, filename, NULL);
	if (utf8Filename == NULL)
	{
		throwByName(env, "java/lang/Exception", "Error getting UTF-8 copy of filename");
		return NULL;
	}
	wcFilenameLen = MultiByteToWideChar(CP_UTF8, 0, utf8Filename, -1, NULL, 0);
	if (wcFilenameLen == 0)
	{
		handleMultiByteToWideCharError(env);
		goto freeMemory;
	}
	wcFilename = malloc(sizeof(*wcFilename) * wcFilenameLen);
	if (wcFilename == NULL)
	{
		throwByName(env, "java/lang/OutOfMemoryError", "Unable to allocate enough memory to convert filename to widechars");
		goto freeMemory;
	}
	if (MultiByteToWideChar(CP_UTF8, 0, utf8Filename, -1, wcFilename, wcFilenameLen) == 0)
	{
		handleMultiByteToWideCharError(env);
		goto freeMemory;
	}

	fileHandle = CreateFileW(wcFilename, GENERIC_READ | (allowWrite ? GENERIC_WRITE : 0), FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL | (syncWriteData ? FILE_FLAG_WRITE_THROUGH : 0) | (sectorMode ? FILE_FLAG_NO_BUFFERING : 0), NULL);
	if (fileHandle == INVALID_HANDLE_VALUE)
	{
		throwByName(env, "java/lang/Exception", "Unable to open file");
		goto freeMemory;
	}

	returnHandle = (*env)->NewByteArray(env, sizeof(returnHandle));
	if (returnHandle == NULL)
	{
		throwByName(env, "java/lang/OutOfMemoryError", "Unable to allocate memory for file handle");
		goto freeMemory;
	}
	(*env)->SetByteArrayRegion(env, returnHandle, 0, sizeof(returnHandle), (jbyte*)&fileHandle);

freeMemory:
	(*env)->ReleaseStringUTFChars(env, filename, utf8Filename);
	if (wcFilename != NULL)
		free(wcFilename);

	return returnHandle;
}

// Close the file
JNIEXPORT void JNICALL Java_org_blinkenbyte_io_WindowsBlockDevice_nativeClose(JNIEnv* env, jclass cls, jbyteArray handleBytes)
{
	HANDLE fileHandle = getHandle(env, handleBytes);
	if (!fileHandle)
		return;
	if (!CloseHandle(fileHandle))
		throwByName(env, "java/io/IOException", "Error calling CloseHandle");
	return;
}

// Sets the file pointer using the specified mode
// Returns the resulting 64-bit file pointer
// If an error occurs, throws an IOException using the specified message
jlong setFilePointer(JNIEnv* env, jclass cls, jbyteArray handleBytes, DWORD moveMethod, jlong jpos, const char* errorMsg)
{
	HANDLE fileHandle = getHandle(env, handleBytes);
	LARGE_INTEGER requestedPos, curPos;
	requestedPos.QuadPart = (LONGLONG)jpos;
	if (!fileHandle)
		return -1;
	if (!SetFilePointerEx(fileHandle, requestedPos, &curPos, moveMethod))
	{
		throwByName(env, "java/io/IOException", errorMsg);
		return -1;
	}
// printf("setFilePointer=%I64d\n", curPos.QuadPart);
	return (jlong)(curPos.QuadPart);
}

// Returns the current 64-bit file pointer
JNIEXPORT jlong JNICALL Java_org_blinkenbyte_io_WindowsBlockDevice_nativeGetFilePointer(JNIEnv* env, jclass cls, jbyteArray handleBytes)
{
	return setFilePointer(env, cls, handleBytes, FILE_CURRENT, 0, "Unable to get current file pointer");
}

// Returns the 64-bit length of the file/device
JNIEXPORT jlong JNICALL Java_org_blinkenbyte_io_WindowsBlockDevice_nativeLength(JNIEnv* env, jclass cls, jbyteArray handleBytes)
{
	HANDLE fileHandle = getHandle(env, handleBytes);
	LARGE_INTEGER fileSize;
	if (!fileHandle)
		return -1;
	if (!GetFileSizeEx(fileHandle, &fileSize))
	{
		throwByName(env, "java/io/IOException", "Unable to get file size");
		return -1;
	}
	return (jlong)(fileSize.QuadPart);
}

JNIEXPORT jint JNICALL Java_org_blinkenbyte_io_WindowsBlockDevice_nativeRead(JNIEnv* env, jclass cls, jbyteArray handleBytes)
{
	HANDLE fileHandle = getHandle(env, handleBytes);
	unsigned char byteToRead;
	DWORD bytesRead;
	if (!fileHandle)
		return -1;
	;
	if (!ReadFile(fileHandle, &byteToRead, 1, &bytesRead, NULL))
	{
		throwByName(env, "java/io/IOException", "Error reading from file");
		return -1;
	}
	if (bytesRead == 0)
		return -1;
	return (jint)byteToRead;
}

JNIEXPORT jint JNICALL Java_org_blinkenbyte_io_WindowsBlockDevice_nativeReadArray(JNIEnv* env, jclass cls, jbyteArray handleBytes, jbyteArray b, jsize offset, jsize len, jint sectorSizeLog)
{
	HANDLE fileHandle = getHandle(env, handleBytes);
	DWORD bytesRead;
	BYTE* readBuffer;
	BYTE* readPtr;
	size_t sectorSize = 1 << sectorSizeLog;
	if (!fileHandle)
		return -1;
	if (sectorSizeLog > 28)
	{
		throwByName(env, "java/lang/OutOfMemoryError", "Unwilling to allocate memory for read buffer: sector size too large");
		return -1;
	}
	// Block devices may need memory aligned on boundaries the same as the sector size,
	// so allocate enough memory to ensure that we can align the buffer on a sector boundary
	readBuffer = malloc(len + sectorSize - 1);
	if (readBuffer == NULL)
	{
		throwByName(env, "java/lang/OutOfMemoryError", "Unable to allocate memory for read buffer");
		return -1;
	}
	readPtr = readBuffer;
	// Adjust the value of readPtr to be aligned on a sector boundary.
	// This may leave unused bytes at the start of the buffer we malloc'ed earlier
	// LONGLONG used here for 64-bit compat. Need intptr_t
	if (sectorSizeLog > 0)
		readPtr = (BYTE*)((LONGLONG)(readPtr + sectorSize - 1) & ~(LONGLONG)(sectorSize - 1));
	if (!ReadFile(fileHandle, readPtr, len, &bytesRead, NULL))
	{
		throwByName(env, "java/io/IOException", "Unable to read from file");
		bytesRead = -1;
	}
	else
		(*env)->SetByteArrayRegion(env, b, offset, bytesRead, readPtr);
	free(readBuffer);
// printf("readArray requested=%d actual=%d\n", len, bytesRead);
	return bytesRead;
}

JNIEXPORT jlong JNICALL Java_org_blinkenbyte_io_WindowsBlockDevice_nativeSeek(JNIEnv* env, jclass cls, jbyteArray handleBytes, jlong pos)
{
	return setFilePointer(env, cls, handleBytes, FILE_BEGIN, pos, "Unable to set file pointer");
}

JNIEXPORT jint JNICALL Java_org_blinkenbyte_io_WindowsBlockDevice_nativeSkipBytes(JNIEnv* env, jclass cls, jbyteArray handleBytes, jint n)
{
	jlong curPos;
	jlong newPos;
	jlong fileSize;

	if (n <= 0)
		return 0;
	curPos = Java_org_blinkenbyte_io_WindowsBlockDevice_nativeGetFilePointer(env, cls, handleBytes);
	if (curPos < 0)
		return 0;
	newPos = curPos + n;

	fileSize = Java_org_blinkenbyte_io_WindowsBlockDevice_nativeLength(env, cls, handleBytes);
	if (fileSize < 0)
		return 0;
	if (curPos == fileSize)
		return 0;
	if (newPos > fileSize)
		newPos = fileSize;
	newPos = Java_org_blinkenbyte_io_WindowsBlockDevice_nativeSeek(env, cls, handleBytes, newPos);
	if (newPos < 0)
		return 0;
	return (jint)(newPos - curPos);
}

JNIEXPORT void JNICALL Java_org_blinkenbyte_io_WindowsBlockDevice_nativeWriteArray(JNIEnv* env, jclass cls, jbyteArray handleBytes, jbyteArray b, jsize offset, jsize len, jint sectorSizeLog)
{
	HANDLE fileHandle = getHandle(env, handleBytes);
	DWORD remainingBytes = len;
	DWORD bytesWritten;
	BYTE* bytesToWrite;
	BYTE* writePtr;
	size_t sectorSize = 1 << sectorSizeLog;
	if (!fileHandle)
		return;
	if (sectorSizeLog > 28)
	{
		throwByName(env, "java/lang/OutOfMemoryError", "Unwilling to allocate memory for write buffer: sector size too large");
		return;
	}
    // Pad buffer to allow sector alignment (see Java_org_blinkenbyte_io_WindowsBlockDevice_nativeReadArray)
	bytesToWrite = malloc(len + sectorSize - 1);
	if (bytesToWrite == NULL)
	{
		throwByName(env, "java/lang/OutOfMemoryError", "Unable to allocate memory for write buffer");
		return;
	}
	writePtr = bytesToWrite;
	// Align buffer on sector boundary (see Java_org_blinkenbyte_io_WindowsBlockDevice_nativeReadArray)
	// LONGLONG used here for 64-bit compat. Need intptr_t
	if (sectorSizeLog > 0)
		writePtr = (BYTE*)((LONGLONG)(writePtr + sectorSize - 1) & ~(LONGLONG)(sectorSize - 1));
	(*env)->GetByteArrayRegion(env, b, offset, len, writePtr);
	if ((*env)->ExceptionOccurred(env) != NULL)
		goto exitWriteArray;
	offset = 0;
	while (remainingBytes > 0)
	{
		if (!WriteFile(fileHandle, writePtr + offset, remainingBytes, &bytesWritten, NULL))
		{
			throwByName(env, "java/io/IOException", "Error writing to the file");
			goto exitWriteArray;
		}
		if (bytesWritten == 0)
		{
			throwByName(env, "java/io/IOException", "Unable to write any bytes to the file");
			goto exitWriteArray;
		}
		remainingBytes -= bytesWritten;
		offset += bytesWritten;
	}
exitWriteArray:
	free(bytesToWrite);
	return;
}

JNIEXPORT void JNICALL Java_org_blinkenbyte_io_WindowsBlockDevice_nativeWrite(JNIEnv* env, jclass cls, jbyteArray handleBytes, jint b)
{
	HANDLE fileHandle = getHandle(env, handleBytes);
	DWORD bytesWritten;
	char c = (char)b;
	if (!fileHandle)
		return;
	if (!WriteFile(fileHandle, &c, 1, &bytesWritten, NULL))
	{
		throwByName(env, "java/io/IOException", "Error writing to the file");
		return;
	}
	if (bytesWritten == 0)
	{
		throwByName(env, "java/io/IOException", "Unable to write byte to the file");
		return;
	}
	return;
}

JNIEXPORT jobject JNICALL Java_org_blinkenbyte_io_WindowsBlockDevice_nativeGetDriveGeometry(JNIEnv* env, jclass cls, jbyteArray handleBytes)
{
	HANDLE fileHandle = getHandle(env, handleBytes);
	DISK_GEOMETRY geometry;
	PARTITION_INFORMATION partitionInfo;
	int outputSize;
	jclass classRef;
	jmethodID classConstructor;

	if (!fileHandle)
		return NULL;
	if (!DeviceIoControl(fileHandle, IOCTL_DISK_GET_DRIVE_GEOMETRY, NULL, 0, &geometry, sizeof(geometry), &outputSize, NULL) || (outputSize != sizeof(geometry)))
	{
		throwByName(env, "java/io/IOException", "Unable to get drive geometry");
		return NULL;
	}
	if (!DeviceIoControl(fileHandle, IOCTL_DISK_GET_PARTITION_INFO, NULL, 0, &partitionInfo, sizeof(partitionInfo), &outputSize, NULL) || (outputSize != sizeof(partitionInfo)))
	{
		throwByName(env, "java/io/IOException", "Unable to get disk size information");
		return NULL;
	}

	if ((classRef = (*env)->FindClass(env, "org/blinkenbyte/io/BlockDevice$DriveGeometry")) == NULL)
		return NULL;
	if ((classConstructor = (*env)->GetMethodID(env, classRef, "<init>", "(JJIII)V")) == NULL)
		return NULL;
	return (*env)->NewObject(env, classRef, classConstructor, (jlong)partitionInfo.PartitionLength.QuadPart, (jlong)geometry.Cylinders.QuadPart, (jint)geometry.TracksPerCylinder, (jint)geometry.SectorsPerTrack, (jint)geometry.BytesPerSector);
}
