// Copyright (c) 2007 Paul Miner <$firstname.$lastname@gmail.com>
// This software is provided 'as-is', without any express or implied warranty. In no event will the authors be held liable for any damages arising from the use of this software.
// Permission is granted to anyone to use this software for any purpose, including commercial applications, and to alter it and redistribute it freely, subject to the following restrictions:
//     1. The origin of this software must not be misrepresented; you must not claim that you wrote the original software. If you use this software in a product, an acknowledgment in the product documentation would be appreciated but is not required.
//     2. Altered source versions must be plainly marked as such, and must not be misrepresented as being the original software.
//     3. This notice may not be removed or altered from any source distribution.

package org.blinkenbyte.io;

import java.io.IOException;
import java.io.Reader;
import java.util.ArrayList;
import java.util.Iterator;

/**
 * A BadBlockEmulator is a device filter that can filter out read/write requests to particular sectors, throwing a BadBlockException instead.
 * The thrown exception details the filtered block that prevented the I/O from completing successfully.
 */
public class BadBlockEmulator extends BlockDeviceFilter
{
  private ArrayList<BlockRange> blockRanges = new ArrayList<BlockRange>();
  
  private boolean allowPartials = true;
  
  /**
   * Constructs a new BadBlockEmulator wrapped around an existing BlockDevice.
   */
  public BadBlockEmulator(BlockDevice _bd)
  {
    super(_bd);
  }
  
  /**
   * Returns whether reads and writes that overlap bad blocks but start with good blocks should be allowed. If false,
   * the entire read or write will fail before the underlying BlockDevice is called. If true (the default), the good blocks
   * from the start of the read/write up to but not including the first bad block will be read or written. If a read is being
   * performed, the read will end there. If a write is being performed, an exception will then be thrown.
   */
  public boolean getAllowPartials()
  {  return allowPartials;  }
  
  /**
   * Sets whether partial reads and writes are allowed. See getAllowPartials() for details.
   */
  public void setAllowPartials(boolean _allowPartials)
  {
    allowPartials = _allowPartials;
    return;
  }
  
  /**
   * Loads a list of block ranges from a Reader, appending them to the existing list.
   */
  public void loadList(Reader r) throws Exception
  {
    BlockRange.readList(r, blockRanges);
    BlockRange.coalesce(blockRanges);
    return;
  }
  
  /**
   * Appends a list of blocks to the current list.
   */
  public void loadList(ArrayList<BlockRange> listToAdd)
  {
    blockRanges.addAll(listToAdd);
    BlockRange.coalesce(blockRanges);
    return;
  }
  
  /**
   * Clears the list of blocks.
   */
  public void clearList()
  {
    blockRanges.clear();
    return;
  }
  
  /**
   * Returns an iterator for the blocks.
   */
  public Iterator<BlockRange> iterator()
  {  return blockRanges.iterator();  }
  
  /**
   * Returns the number of BlockRange objects contained in this object.
   */
  public int numBlocks()
  {  return blockRanges.size();  }
  
  /**
   * Returns the BlockRange at index <i>index</i> within this object.
   */
  public BlockRange getBlockRange(int index)
  {  return blockRanges.get(index);  }
  
  /**
   * Adds a new range of blocks to this object.
   *
   * @param blockNumber The starting block number of the range.
   * @param blockLength The length of the range.
   */
  public void addBlock(long blockNumber, long blockLength) throws Exception
  {
    blockRanges.add(new BlockRange(blockNumber, blockLength));
    BlockRange.coalesce(blockRanges);
    return;
  }
  
  /**
   * Removes a range of blocks from this object.
   *
   * @param blockNumber The starting block number of the range.
   * @param blockLength The length of the range.
   */
  public boolean removeBlock(long blockNumber, long blockLength) throws Exception
  {
    int numBlocks = blockRanges.size();
    if (numBlocks == 0)
      return false;
    BlockRange blockToRemove = new BlockRange(blockNumber, blockLength);
    int curIndex = getInsertionIndex(blockNumber);
    if (curIndex > 0)
      if (blockRanges.get(curIndex - 1).isBlockInRange(blockNumber))
        curIndex--;
    boolean removedSomething = false;
    while (curIndex < blockRanges.size())
    {
      BlockRange curBlock = blockRanges.get(curIndex);
      if (!curBlock.isOverlapping(blockToRemove))
        break;
      removedSomething = true;
      BlockRange[] remainingPieces = curBlock.remove(blockToRemove);
      if (remainingPieces == null)
      {
        blockRanges.remove(curIndex);
        continue;
      }
      blockRanges.set(curIndex++, remainingPieces[0]);
      if (remainingPieces.length == 2)
      {
        blockRanges.add(curIndex, remainingPieces[1]);
        break;
      }
    }
    return removedSomething;
  }
  
  private int getInsertionIndex(long blockNumber)
  {
    int low = 0, mid = 0, high = blockRanges.size();
    while (low < high)
    {
      mid = (low + high) >>> 1;
      BlockRange curPos = blockRanges.get(mid);
      if (curPos.getBlockNumber() < blockNumber)
        low = mid + 1;
      else
      {
        if (curPos.getBlockNumber() > blockNumber)
          high = mid;
        else
          return mid;
      }
    }
    return low;
  }
  
  private int getMatchingBlockIndex(long blockNumber)
  {
    int ind = getInsertionIndex(blockNumber);
    if (ind < blockRanges.size())
      if (blockRanges.get(ind).isBlockInRange(blockNumber))
        return ind;
    if (ind > 0)
      if (blockRanges.get(--ind).isBlockInRange(blockNumber))
        return ind;
    return -1;
  }
  
  private int getFirstOverlapIndex(BlockRange bp)
  {
    int ind = getInsertionIndex(bp.getBlockNumber());
    if (ind > 0)
      if (blockRanges.get(ind - 1).isBlockInRange(bp.getBlockNumber()))
        return ind - 1;
    if (ind < blockRanges.size())
      if (bp.isBlockInRange(blockRanges.get(ind).getBlockNumber()))
        return ind;
    return -1;
  }
  
  /**
   * Reads a byte from the underlying block device, throwing an exception if the byte is in a bad block range.
   */
  public int read() throws IOException
  {
    long curPos = bd.getFilePointer();
    long blockNumber = curPos >>> getSectorSizeLog();
    int ind = getMatchingBlockIndex(blockNumber);
    if (ind < 0)
      return bd.read();
    throw new BadBlockException("read() @" + curPos + " (block #" + blockNumber + ") matches bad block: " + blockRanges.get(ind).toString());
  }
  
  /**
   * Reads up to <i>len</i> bytes from the underlying block device into a byte array, starting at <i>b[offset]</i>. If the bytes to be read
   * do not overlap any bad blocks, the read proceeds normally. If the start of the read overlaps any bad blocks, the read fails immediately.
   * If the bad blocks are not the first bytes to be read, the result depends on whether partial reads are allowed. If partial reads are
   * allowed, a read request for the maximum possible number of bytes that could be read before reading a bad block is processed and returned.
   * If partials are not allowed, the read fails immediately.
   */
  public int read(byte[] b, int offset, int len) throws IOException
  {
    long curPos = bd.getFilePointer();
    long blockNumber = curPos >>> getSectorSizeLog();
    long endBlockNumber = (curPos + len - 1) >>> getSectorSizeLog();
    BlockRange bp = new BlockRange(blockNumber, endBlockNumber - blockNumber + 1);
    int overlapIndex = getFirstOverlapIndex(bp);
    if (overlapIndex < 0)
      return bd.read(b, offset, len);
    BlockRange overlappingBlock = blockRanges.get(overlapIndex);
    if (allowPartials)
    {
      long badBlockOffset = overlappingBlock.getBlockNumber() << getSectorSizeLog();
      if (badBlockOffset > curPos)
        return bd.read(b, offset, (int)(badBlockOffset - curPos));
    }
    throw new BadBlockException("read() of " + len + " bytes @" + curPos + " overlaps bad block: " + overlappingBlock.toString());
  }

  /**
   * Writes <i>len</i> bytes to the underlying block device from a byte array, starting at <i>b[offset]</i>. If the bytes to be written
   * do not overlap any bad blocks, the write proceeds normally. If the start of the write overlaps any bad blocks, the write fails immediately.
   * If the bad blocks are not the first bytes to be written, the result depends on whether partial writes are allowed. If partial writes are
   * allowed, all bytes up to the start of the bad block range are written, after which an exception is thrown. If partials writes are not
   * allowed, the write fails immediately.
   */
  public void write(byte[] b, int offset, int len) throws IOException
  {
    long curPos = bd.getFilePointer();
    long blockNumber = curPos >>> getSectorSizeLog();
    long endBlockNumber = (curPos + len - 1) >>> getSectorSizeLog();
    BlockRange bp = new BlockRange(blockNumber, endBlockNumber - blockNumber + 1);
    int overlapIndex = getFirstOverlapIndex(bp);
    if (overlapIndex < 0)
    {
      bd.write(b, offset, len);
      return;
    }
    //
    BlockRange overlappingBlock = blockRanges.get(overlapIndex);
    int bytesToWrite = 0;
    if (allowPartials)
    {
      long badBlockOffset = overlappingBlock.getBlockNumber() << getSectorSizeLog();
      if (badBlockOffset > curPos)
      {
        bytesToWrite = (int)(badBlockOffset - curPos);
        bd.write(b, offset, bytesToWrite);
      }
    }
    throw new BadBlockException("write() of " + len + " bytes @" + curPos + " overlaps bad block" + ((bytesToWrite > 0) ? (" (but was able to write " + bytesToWrite + " bytes)") : "") + ": " + overlappingBlock.toString());
  }
  
  /**
   * Writes a byte to the underlying block device, throwing an exception if the byte is in a bad block range.
   */
  public void write(int b) throws IOException
  {
    long curPos = bd.getFilePointer();
    long blockNumber = curPos >>> getSectorSizeLog();
    int ind = getMatchingBlockIndex(blockNumber);
    if (ind < 0)
    {
      bd.write(b);
      return;
    }
    throw new BadBlockException("write() @" + curPos + " (block #" + blockNumber + ") matches bad block: " + blockRanges.get(ind).toString());
  }
  
  /**
   * An exception class that is thrown when bad blocks are requested to be read or written by BadBlockEmulator.
   */
  public static class BadBlockException extends IOException
  {
    public BadBlockException(String msg)
    {  super(msg);  }
  }

}
