// Copyright (c) 2007 Paul Miner <$firstname.$lastname@gmail.com>
// This software is provided 'as-is', without any express or implied warranty. In no event will the authors be held liable for any damages arising from the use of this software.
// Permission is granted to anyone to use this software for any purpose, including commercial applications, and to alter it and redistribute it freely, subject to the following restrictions:
//     1. The origin of this software must not be misrepresented; you must not claim that you wrote the original software. If you use this software in a product, an acknowledgment in the product documentation would be appreciated but is not required.
//     2. Altered source versions must be plainly marked as such, and must not be misrepresented as being the original software.
//     3. This notice may not be removed or altered from any source distribution.

package org.blinkenbyte.io;

import java.io.IOException;

/**
 * This interface models a block device, providing methods for random reads and writes, and querying device geometry.
 */
public interface BlockDevice
{

  /**
   * Close the device.
   */
  public void close() throws IOException;
  
  /**
   * Open the device.
   *
   * @param filename          The filename or device name to open
   * @param allowWrite        If true, allow write access to the device.
   * @param syncWriteData     If true, always flush writes to the device.
   * @param syncWriteMetaData If true, always flush writes of the device's metadata.
   * @param sectorMode        If true, all reads and writes should be on sector boundaries,
   *                          and exceptions should be thrown if a request is not aligned.
   */
  public void open(String filename, boolean allowWrite, boolean syncWriteData, boolean syncWriteMetaData, boolean sectorMode) throws Exception;
  
  /**
   * Returns the current pointer for read and write operations.
   */
  public long getFilePointer() throws IOException;
  
  /**
   * Returns the size in bytes of the device.
   */
  public long length() throws IOException;
  
  /**
   * Reads and returns a single byte.
   */
  public int read() throws IOException;
  
  /**
   * Reads bytes into the array. Unless the end of the device has been reached, at least one byte will be read, and at most <i>b.length</i> bytes.
   */
  public int read(byte[] b) throws IOException;

  /**
   * Reads bytes into the array. Unless the end of the device has been reached, at least one byte will be read, and at most <i>len</i> bytes.
   */
  public int read(byte[] b, int offset, int len) throws IOException;
  
  /**
   * Seeks the read/write pointer to the specified position. Position is 0-based.
   */
  public void seek(long pos) throws IOException;
  
  /**
   * Advances the read/write pointer <i>n</i> bytes.
   */
  public int skipBytes(int n) throws IOException;
  
  /**
   * Writes the contents of <i>b</i> to the device.
   */
  public void write(byte[] b) throws IOException;
  
  /**
   * Writes <i>len</i> bytes to the device starting from <i>b[offset]</i>.
   */
  public void write(byte[] b, int offset, int len) throws IOException;
  
  /**
   * Writes one byte to the device.
   */
  public void write(int b) throws IOException;
  
  /**
   * Returns an object describing the geometry of the device.
   */
  public BlockDevice.DriveGeometry getDriveGeometry() throws Exception;
  
  /**
   * Returns the sector size of the device.
   */
  public int getSectorSize();
  
  /**
   * Sets the sector size of the device.
   */
  public void setSectorSize(int sectorSize) throws Exception;
  
  /**
   * Returns the base 2 logarithm of the sector size of the device.
   */
  public int getSectorSizeLog();
  
  /**
   * Returns whether the device is set to automatically align reads/writes to sector boundaries. If an unaligned read is performed,
   * an aligned read should be performed and the relevant part of the data copied to the buffer. If an unaligned write is performed,
   * an aligned read should be performed to get the values of the bytes not being written, after which an aligned write should be performed.
   */
  public boolean getAutoSectorAlign();
  
  /**
   * Sets whether the device should automatically align read/write requests.
   */
  public void setAutoSectorAlign(boolean _autoSectorAlign);

  /**
   * Class representing the geometry of a device
   */
  public static class DriveGeometry
  {
    /**
     * Size of the device in bytes.
     */
    public long diskSize = -1L;
    
    /**
     * Number of tracks or cylinders.
     */
    public long numCylinders = -1L;
    
    /**
     * Number of heads.
     */
    public int numHeads = -1;
    
    /**
     * Number of sectors per track.
     */
    public int numSectorsPerTrack = -1;
    
    /**
     * Number of bytes per sector.
     */
    public int numBytesPerSector = -1;
    
    public DriveGeometry()
    {  }
    
    public DriveGeometry(long _diskSize, long _numCylinders, int _numHeads, int _numSectorsPerTrack, int _numBytesPerSector)
    {
      diskSize = _diskSize;
      numCylinders = _numCylinders;
      numHeads = _numHeads;
      numSectorsPerTrack = _numSectorsPerTrack;
      numBytesPerSector = _numBytesPerSector;
    }
  }
}
