// Copyright (c) 2007 Paul Miner <$firstname.$lastname@gmail.com>
// This software is provided 'as-is', without any express or implied warranty. In no event will the authors be held liable for any damages arising from the use of this software.
// Permission is granted to anyone to use this software for any purpose, including commercial applications, and to alter it and redistribute it freely, subject to the following restrictions:
//     1. The origin of this software must not be misrepresented; you must not claim that you wrote the original software. If you use this software in a product, an acknowledgment in the product documentation would be appreciated but is not required.
//     2. Altered source versions must be plainly marked as such, and must not be misrepresented as being the original software.
//     3. This notice may not be removed or altered from any source distribution.

package org.blinkenbyte.io;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.IOException;
import java.io.Reader;
import java.io.Writer;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

/**
 * This class represents a mutable block range, specified by a starting block number, and the length of the run. It provides
 * methods to combine and remove blocks individually and as a list, and to read and write lists from a Reader/Writer.
 */
public class BlockRange implements Comparable<BlockRange>
{
  private long blockNumber = 0L;
  private long blockLength = 0L;
  
  /**
   * Constructs a new BlockRange.
   *
   * @param _blockNumber The starting block number.
   * @param _blockLength The number of blocks.
   */
  public BlockRange(long _blockNumber, long _blockLength)
  {
    blockNumber = _blockNumber;
    blockLength = _blockLength;
  }
  
  /**
   * Returns the starting block number represented by this object.
   */
  public long getBlockNumber()
  {  return blockNumber;  }
  
  /**
   * Returns the number of blocks represented by this object.
   */
  public long getBlockLength()
  {  return blockLength;  }
  
  /**
   * Returns the ending block number encompassed by this range. This is one block past the last block covered by this object.
   */
  public long getEndBlockNumber()
  {
    return blockNumber + blockLength;
  }
  
  /**
   * Returns whether a given block number is in the range covered by this object.
   */
  public boolean isBlockInRange(long n)
  {
    return (compareToBlockNumber(n) == 0);
  }
  
  /**
   * Compares a block number to this range.
   *
   * @return Returns 1 if the block number is past the range of this object.<br>
   *         Returns 0 if the block number is within the range of this object.<br>
   *         Returns -1 if the block number is before the range of this object.
   */
  public int compareToBlockNumber(long n)
  {
    if (blockNumber > n)
      return 1;
    if (getEndBlockNumber() <= n)
      return -1;
    return 0;
  }
  
  /**
   * Returns true if the specified BlockRange is fully contained within this BlockRange.
   */
  public boolean contains(BlockRange bp)
  {
    return (bp.blockNumber >= blockNumber) && (bp.getEndBlockNumber() <= getEndBlockNumber());
  }
  
  /**
   * Returns true if the union of this BlockRange with the specified BlockRange is a single continuous block.
   */
  public boolean isContiguousWith(BlockRange bp)
  {
    if (blockNumber < bp.blockNumber)
      return (getEndBlockNumber() >= bp.blockNumber);
    return (bp.getEndBlockNumber() >= blockNumber);
  }
  
  /**
   * Returns true if the intersection of this BlockRange with the specified BlockRange is not empty.
   */
  public boolean isOverlapping(BlockRange bp)
  {
    if (blockNumber < bp.blockNumber)
      return (getEndBlockNumber() > bp.blockNumber);
    return (bp.getEndBlockNumber() > blockNumber);
  }
  
  /**
   * Extends the range of this BlockRange to include the specified block.
   */
  public void absorb(BlockRange bp)
  {
    long endBlockNumber = getEndBlockNumber();
    long bpEndBlockNumber = bp.getEndBlockNumber();
    blockNumber = (blockNumber < bp.blockNumber) ? blockNumber : bp.blockNumber;
    blockLength = ((endBlockNumber > bpEndBlockNumber) ? endBlockNumber : bpEndBlockNumber) - blockNumber;
    return;
  }
  
  /**
   * Returns a new block that represents a union of this BlockRange and the specified BlockRange. If the blocks are not contiguous, all blocks
   * between the two blocks are included in the union.
   */
  public BlockRange union(BlockRange bp)
  {
    long endBlockNumber = getEndBlockNumber();
    long bpEndBlockNumber = bp.getEndBlockNumber();
    long lowestStart = (blockNumber < bp.blockNumber) ? blockNumber : bp.blockNumber;
    long largestEnd = (endBlockNumber > bpEndBlockNumber) ? endBlockNumber : bpEndBlockNumber;
    return new BlockRange(lowestStart, largestEnd - lowestStart);
  }
  
  /**
   * Removes a range of blocks from this BlockRange, and returns the resulting BlockRange(s). If the ranges do not overlap, the result is undefined.
   *
   * @return Returns the result of the block removal. This may be a null if the specified BlockRange contains this BlockRange. Otherwise, an array
   * of one or two BlockRange objects will be returned.
   */
  public BlockRange[] remove(BlockRange bp)
  {
    long endBlockNumber = getEndBlockNumber();
    long bpEndBlockNumber = bp.getEndBlockNumber();
    BlockRange firstPiece = null;
    BlockRange secondPiece = null;
    if (blockNumber < bp.blockNumber)
    {
      firstPiece = new BlockRange(blockNumber, bp.blockNumber - blockNumber);
      if (endBlockNumber <= bpEndBlockNumber)
        return new BlockRange[] { firstPiece };
      secondPiece = new BlockRange(bpEndBlockNumber, endBlockNumber - bpEndBlockNumber);
      return new BlockRange[] { firstPiece, secondPiece };
    }
    if (endBlockNumber <= bpEndBlockNumber)
      return null;
    secondPiece = new BlockRange(bpEndBlockNumber, endBlockNumber - bpEndBlockNumber);
    return new BlockRange[] { secondPiece };
  }
  
  /**
   * Compares the starting block of this BlockRange to the specified BlockRange.
   */
  public int compareTo(BlockRange bp)
  {
    if (blockNumber < bp.blockNumber)
      return -1;
    if (blockNumber > bp.blockNumber)
      return 1;
    return 0;
  }
  
  /**
   * Returns a string representation of this BlockRange
   */
  public String toString()
  {
    if (blockLength <= 0)
      return "Block #" + blockNumber + " (no blocks)";
    if (blockLength == 1)
      return "Block #" + blockNumber + " (1 block)";
    return "Blocks #" + blockNumber + "-" + (blockNumber + blockLength - 1) + " (" + blockLength + " blocks)";
  }
  
  /**
   * Returns a string representation of this BlockRange as it would be read or written from a Reader/Writer.
   * This is the block number, followed by a space, and followed by a block count.
   */
  public String getLine()
  {  return String.valueOf(blockNumber) + " " + String.valueOf(blockLength);  }
  
  /**
   * Coalesces a list of BlockRange objects. Any contiguous or overlapping blocks will be combined. This method
   * may modify existing ranges rather than create new ones.
   */
  public static void coalesce(List<BlockRange> list)
  {
    Collections.sort(list);
    if (list.size() > 1)
    {
      int newSize = 1;
      BlockRange curBlock = list.get(0);
      for (int i = 1; i < list.size(); i++)
      {
        BlockRange bp = list.get(i);
        if (curBlock.isContiguousWith(bp))
          curBlock.absorb(bp);
        else
        {
          if (i != newSize)
            list.set(newSize, bp);
          curBlock = bp;
          newSize++;
        }
      }
      for (int i = list.size() - 1; i >= newSize; i--)
        list.remove(i);
    }
    return;
  }
  
  /**
   * Writes out a list of blocks in a format that can be read by readList. One BlockRange is written per line, using
   * the result of getLine().
   */
  public static void writeList(List<BlockRange> list, Writer w) throws IOException
  {
    BufferedWriter bw = null;
    if (w instanceof BufferedWriter)
      bw = (BufferedWriter)w;
    else
      bw = new BufferedWriter(w);
    for (BlockRange range : list)
    {
      bw.write(range.getLine());
      bw.newLine();
    }
    bw.flush();
    return;
  }
  
  /**
   * Reads a list of block ranges in a format that can be read by readList. One BlockRange is written per line, using
   * the result of getLine().
   *
   * @return A list of BlockRange objects
   */
  public static ArrayList<BlockRange> readList(Reader r) throws Exception
  {
    ArrayList<BlockRange> list = new ArrayList<BlockRange>();
    readList(r, list);
    return list;
  }
  
  /**
   * Reads a list of block ranges in a format that can be read by readList, and appends them to the specified list. One BlockRange
   * is written per line, using the result of getLine().
   */
  public static void readList(Reader r, List<BlockRange> list) throws Exception
  {
    BufferedReader br = null;
    if (br instanceof BufferedReader)
      br = (BufferedReader)r;
    else
      br = new BufferedReader(r);
    String line = null;
    long[] nums = new long[2];
    while ((line = br.readLine()) != null)
    {
      line = line.trim();
      int lineLen = line.length();
      int startIndex = -1, numIndex = 0;
      for (int i = 0; i <= lineLen; i++)
      {
        char c = (i == lineLen) ? (char)0 : line.charAt(i);
        if (((c >= '0') && (c <= '9')) || ((c >= 'A') && (c <= 'F')) || ((c >= 'a') && (c <= 'f')) || (c == 'x') || (c == 'X'))
        {
          if (startIndex < 0)
          {
            if (numIndex == 2)
              throw new Exception("Extra characters on line");
            startIndex = i;
          }
        }
        else
        {
          if ((c != ' ') && (c != '\t') && (c != ',') && (i != lineLen))
            throw new Exception("Input contains invalid characters");
          if (startIndex >= 0)
          {
            String num = line.substring(startIndex, i);
            startIndex = -1;
            if (num.startsWith("0x") || num.startsWith("0X"))
              nums[numIndex] = Long.parseLong(num.substring(2), 16);
            else
              nums[numIndex] = Long.parseLong(num);
            numIndex++;
          }
        }
      }
      if (numIndex == 0)
        continue;
      if (numIndex == 1)
        list.add(new BlockRange(nums[0], 1L));
      else
        list.add(new BlockRange(nums[0], nums[1]));
    }
    return;
  }
  
}
