// Copyright (c) 2007 Paul Miner <$firstname.$lastname@gmail.com>
// This software is provided 'as-is', without any express or implied warranty. In no event will the authors be held liable for any damages arising from the use of this software.
// Permission is granted to anyone to use this software for any purpose, including commercial applications, and to alter it and redistribute it freely, subject to the following restrictions:
//     1. The origin of this software must not be misrepresented; you must not claim that you wrote the original software. If you use this software in a product, an acknowledgment in the product documentation would be appreciated but is not required.
//     2. Altered source versions must be plainly marked as such, and must not be misrepresented as being the original software.
//     3. This notice may not be removed or altered from any source distribution.

package org.blinkenbyte.io;

import java.io.InputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.util.Arrays;

/**
 * This class represents an output stream that uses two-dimensional byte arrays for storage. This allows it to allocate
 * storage space incrementally without the usual create/copy cycle usually needed to expand an array. It also supports
 * creating an InputStream that can read directly from the byte arrays.
 */
public class ByteArrayBuffer extends OutputStream
{
  /**
   * The base 2 logarithm of the size of the sub-buffers.
   */
  protected int SUB_BUFFER_LOG = 16;

  /**
   * The size of the sub-buffers.
   */
  protected int SUB_BUFFER_SIZE = 0;

  /**
   * The bitmask used to break an index into a main index and a sub index.
   */
  protected int SUB_BUFFER_MASK = 0;

  /**
   * The size of the main buffer.
   */
  protected int MAIN_BUFFER_SIZE = 256;

  /**
   * The main buffer.
   */
  protected byte[][] mainBuffer = null;

  /**
   * The number of bytes currently used.
   */
  protected int bytesUsed = 0;

  /**
   * The current location of the write pointer.
   */
  protected int writePtr = 0;

  /**
   * The number of buffers allocated in the main buffer.
   */
  protected int mainBuffersAllocated = 0;
  
  /**
   * Constructs a new ByteArrayBuffer using default settings (64KB sub-buffers)
   */
  public ByteArrayBuffer()
  {
    initBuffer();
  }
  
  /**
   * Constructs a new ByteArrayBuffer with the specified dimensions.
   *
   * @param mainBufferSize The size of the main buffer. The main buffer simply holds sub-buffers, and will automatically grow as needed.
   * @param subBufferLog   The base 2 logarithm of the size of the sub-buffer. Sub-buffers are allocated at their full size as needed.
   */
  public ByteArrayBuffer(int mainBufferSize, int subBufferLog)
  {
    MAIN_BUFFER_SIZE = mainBufferSize;
    SUB_BUFFER_LOG = subBufferLog;
    initBuffer();
  }
  
  private void initBuffer()
  {
    SUB_BUFFER_SIZE = 1 << SUB_BUFFER_LOG;
    SUB_BUFFER_MASK = SUB_BUFFER_SIZE - 1;
    mainBuffer = new byte[MAIN_BUFFER_SIZE][];
  }
  
  /**
   * Inherited from OutputStream, but does nothing.
   */
  public void close()
  {
    // Do nothing, because data is intended to be reused
    return;
  }
  
  /**
   * Nulls all pointers to aid with garbage collection.
   */
  public void clean()
  {
    bytesUsed = 0;
    writePtr = 0;
    mainBuffersAllocated = 0;
    if (mainBuffer != null)
      for (int i = 0; i < mainBuffer.length; i++)
        mainBuffer[i] = null;
    mainBuffer = null;
    return;
  }
  
  /**
   * Inherited from OutputStream; does nothing.
   */
  public void flush()
  {
    return;
  }
  
  /**
   * Writes the byte array <i>b</i> to the buffers.
   */
  public void write(byte[] b)
  {
    write(b, 0, b.length);
    return;
  }
  
  /**
   * Writes <i>len</i> bytes from a byte array starting at <i>b[offset]</i>.
   */
  public void write(byte[] b, int offset, int len)
  {
    ensureSpace(len);
    while (len > 0)
    {
      int bytesToCopy = SUB_BUFFER_SIZE - (writePtr & SUB_BUFFER_MASK);
      if (bytesToCopy > len)
        bytesToCopy = len;
      System.arraycopy(b, offset, mainBuffer[writePtr >>> SUB_BUFFER_LOG], writePtr & SUB_BUFFER_MASK, bytesToCopy);
      writePtr += bytesToCopy;
      if (writePtr > bytesUsed)
        bytesUsed = writePtr;
      offset += bytesToCopy;
      len -= bytesToCopy;
    }
    return;
  }
  
  /**
   * Writes a single byte to the buffers.
   */
  public void write(int b)
  {
    ensureSpace(1);
    mainBuffer[writePtr >>> SUB_BUFFER_LOG][writePtr & SUB_BUFFER_MASK] = (byte)b;
    writePtr++;
    if (writePtr > bytesUsed)
      bytesUsed = writePtr;
    return;
  }
  
  /**
   * Returns the current number of bytes used in the buffer.
   */
  public int length()
  {
    return bytesUsed;
  }
  
  /**
   * Adjusts the number of bytes used in the buffer. If the length is increasing, all new bytes are initialized to 0.
   */
  public void setLength(int newLength)
  {
    if (bytesUsed == newLength)
      return;
    if (bytesUsed > newLength)
    {
      bytesUsed = newLength;
      if (writePtr > newLength)
        writePtr = newLength;
      return;
    }
    // Increasing length
    // Allocate space, fill with zeroes
    int mainIndex = bytesUsed >>> SUB_BUFFER_LOG;
    int subIndex = bytesUsed & SUB_BUFFER_MASK;
    int bytesNeeded = newLength - bytesUsed;
    int totalBytesToZero = bytesNeeded;
    for (int i = mainIndex; i < mainBuffersAllocated; i++)
    {
      int startIndex = (i == mainIndex) ? subIndex : 0;
      int endIndex = SUB_BUFFER_SIZE;
      if ((startIndex - endIndex) > totalBytesToZero)
        endIndex = startIndex + totalBytesToZero;
      Arrays.fill(mainBuffer[i], startIndex, endIndex, (byte)0);
      totalBytesToZero -= (endIndex - startIndex);
      if (totalBytesToZero == 0)
        break;
    }
    // Anything newly allocated will be zeroes
    ensureSpace(bytesNeeded);
    bytesUsed = newLength;
    return;
  }
  
  /**
   * Changes the write pointer.
   */
  public int seek(int newWritePtr)
  {
    if (newWritePtr < 0)
      writePtr = bytesUsed + newWritePtr;
    else
      writePtr = newWritePtr;
    if (writePtr < 0)
      writePtr = 0;
    else
    {
      if (writePtr > bytesUsed)
      {
        ensureSpace(writePtr - bytesUsed);
        bytesUsed = writePtr;
      }
    }
    return writePtr;
  }
  
  /**
   * Gets the current value of the write pointer
   */
  public int getWritePtr()
  {
    return writePtr;
  }
  
  /**
   * Fills all allocated buffers with 0's.
   */
  public void fillZeroes()
  {
    for (int i = 0; i < mainBuffersAllocated; i++)
    {
      byte[] b = mainBuffer[i];
      Arrays.fill(b, 0, SUB_BUFFER_SIZE, (byte)0);
    }
    return;
  }
  
  private int bytesAvailable()
  {
    int mainIndex = bytesUsed >>> SUB_BUFFER_LOG;
    if (mainIndex >= mainBuffersAllocated)
      return 0;
    return ((mainBuffersAllocated - mainIndex) << SUB_BUFFER_LOG) - (bytesUsed & SUB_BUFFER_MASK);
  }
  
  private void ensureSpace(int n)
  {
    if (n <= 0)
      return;
    if (n >= 1073741824)
      throw new RuntimeException("ensureSpace() called with parameter over 1GB");
    int avail = bytesAvailable();
    if (n <= avail)
      return;
    int numBuffersToAllocate = (n - avail + SUB_BUFFER_MASK) >>> SUB_BUFFER_LOG;
    int totalBuffersNeeded = mainBuffersAllocated + numBuffersToAllocate;
    int newMainBufferSize = mainBuffer.length;
    while (newMainBufferSize < totalBuffersNeeded)
      newMainBufferSize <<= 1;
    if (mainBuffer.length != newMainBufferSize)
    {
      byte[][] newMainBuffer = new byte[newMainBufferSize][];
      if (mainBuffersAllocated > 0)
        System.arraycopy(mainBuffer, 0, newMainBuffer, 0, mainBuffersAllocated);
      mainBuffer = newMainBuffer;
    }
    for (int i = 0; i < numBuffersToAllocate; i++)
      mainBuffer[mainBuffersAllocated + i] = new byte[SUB_BUFFER_SIZE];
    mainBuffersAllocated = totalBuffersNeeded;
    return;
  }
  
  /**
   * Returns a ByteArrayBufferInputStream that can be used to read directly from the internal buffers of this object. The returned stream
   * will have its own read pointer independent of the write pointer.
   */
  public ByteArrayBufferInputStream getInputStream()
  {
    return new ByteArrayBufferInputStream(this);
  }
  
  /**
   * Returns a ByteArrayBufferInputStream that can be used to read directly from the internal buffers of this object. The returned stream
   * can specify an independent or shared read pointer. If it is shared, reads affect the write pointer and writes affect the read pointer.
   *
   * @param _sharedPointer Specifies whether the read pointer should be shared with the write pointer.
   */
  public ByteArrayBufferInputStream getInputStream(boolean _sharedPointer)
  {
    return new ByteArrayBufferInputStream(this, _sharedPointer);
  }
  
  /**
   * Class that can extends InputStream and can read directly from the internal buffers of the host ByteArrayBuffer. This prevents
   * unnecessary duplication of data, and a way of using the buffer as a device that can be read and written from.
   */
  public static class ByteArrayBufferInputStream extends InputStream
  {
    /**
     * The ByteArrayBufffer this InputStream reads from.
     */
    protected ByteArrayBuffer bab = null;
    
    /**
     * The read pointer used by this input stream.
     */
    protected int readPtr = 0;
    
    /**
     * The mark pointer.
     */
    protected int markPtr = -1;
    
    /**
     * The mark read limit.
     */
    protected int markReadLimit = 0;
    
    /**
     * Copy of SUB_BUFFER_LOG from the host ByteArrayBuffer.
     */
    protected int SUB_BUFFER_LOG = 0;

    /**
     * Copy of SUB_BUFFER_SIZE from the host ByteArrayBuffer.
     */
    protected int SUB_BUFFER_SIZE = 0;

    /**
     * Copy of SUB_BUFFER_MASK from the host ByteArrayBuffer.
     */
    protected int SUB_BUFFER_MASK = 0;

    /**
     * Whether this input stream should share a pointer with the host ByteArrayBuffer.
     */
    protected boolean sharedPointer = false;
    
    /**
     * Constructs a new ByteArrayBufferInputStream from an existing ByteArrayBuffer
     */
    public ByteArrayBufferInputStream(ByteArrayBuffer _bab)
    {
      bab = _bab;
      SUB_BUFFER_LOG = bab.SUB_BUFFER_LOG;
      SUB_BUFFER_SIZE = bab.SUB_BUFFER_SIZE;
      SUB_BUFFER_MASK = bab.SUB_BUFFER_MASK;
    }
    
    /**
     * Constructs a new ByteArrayBufferInputStream from an existing ByteArrayBuffer and specifies whether they should share a read/write pointer.
     */
    public ByteArrayBufferInputStream(ByteArrayBuffer _bab, boolean _sharedPointer)
    {
      this(_bab);
      sharedPointer = _sharedPointer;
    }
    
    /**
     * Returns the number of bytes available to be read.
     */
    public int available()
    {
      return bab.length() - (sharedPointer ? bab.writePtr : readPtr);
    }
    
    /**
     * Closes this stream (nulls the reference to the ByteArrayBuffer).
     */
    public void close()
    {
      bab = null;
      return;
    }
    
    /**
     * Sets the limit for mark().
     */
    public void mark(int readLimit)
    {
      markReadLimit = readLimit;
      return;
    }
    
    /**
     * Returns true.
     */
    public boolean markSupported()
    {
      return true;
    }
    
    /**
     * Reads a single byte.
     */
    public int read()
    {
      if (sharedPointer)
        readPtr = bab.writePtr;
      if (readPtr >= bab.length())
        return -1;
      int b = (int)bab.mainBuffer[readPtr >>> SUB_BUFFER_LOG][readPtr & SUB_BUFFER_MASK];
      if (sharedPointer)
        bab.writePtr++;
      else
        readPtr++;
      return b & 0xFF;
    }
    
    /**
     * Reads up to <i>b.length</i> bytes into <i>b</i>.
     */
    public int read(byte[] b)
    {
      return read(b, 0, b.length);
    }
    
    /**
     * Reads up to <i>offset</i> bytes into array <i>b</i> starting at <i>b[offset]</i>.
     */
    public int read(byte[] b, int offset, int len)
    {
      int originalLen = len;
      int babLen = bab.length();
      if (sharedPointer)
        readPtr = bab.writePtr;
      while ((len > 0) && (readPtr < babLen))
      {
        int bytesToCopy = SUB_BUFFER_SIZE - (readPtr & SUB_BUFFER_MASK);
        if (bytesToCopy > len)
          bytesToCopy = len;
        if ((readPtr + bytesToCopy) > babLen)
          bytesToCopy = babLen - readPtr;
        System.arraycopy(bab.mainBuffer[readPtr >>> SUB_BUFFER_LOG], readPtr & SUB_BUFFER_MASK, b, offset, bytesToCopy);
        if (sharedPointer)
          bab.writePtr += bytesToCopy;
        readPtr += bytesToCopy;
        offset += bytesToCopy;
        len -= bytesToCopy;
      }
      return originalLen - len;
    }
    
    /**
     * Resets the read pointer back to the mark. If no mark has been set or too many bytes have been read since the mark,
     * an IOException is thrown.
     */
    public void reset() throws IOException
    {
      if (markPtr < 0)
        throw new IOException("No mark to reset() to");
      if ((sharedPointer ? bab.writePtr : readPtr) > (markPtr + markReadLimit))
        throw new IOException("Too many bytes read since mark()");
      if (sharedPointer)
        bab.writePtr = markPtr;
      else
        readPtr = markPtr;
      return;
    }
    
    /**
     * Advances the pointer <i>n</i> bytes.
     */
    public long skip(long n)
    {
      if (n <= 0L)
        return 0L;
      long newReadPtr = 0L;
      if (sharedPointer)
        readPtr = bab.writePtr;
      if ((n >= Integer.MAX_VALUE) || ((newReadPtr = readPtr + n) >= bab.length()))
      {
        int bytesSkipped = bab.length() - readPtr;
        readPtr = bab.length();
        if (sharedPointer)
          bab.writePtr = readPtr;
        return bytesSkipped;
      }
      readPtr = (int)newReadPtr;
      if (sharedPointer)
        bab.writePtr = readPtr;
      return n;
    }

    /**
     * Seeks the pointer to <i>newReadPtr</i>.
     */
    public int seek(int newReadPtr)
    {
      int len = bab.length();
      if (newReadPtr < 0)
        readPtr = len + newReadPtr;
      else
        readPtr = newReadPtr;
      if (readPtr < 0)
        readPtr = 0;
      else
        if (readPtr > len)
          readPtr = len;
      if (sharedPointer)
        bab.writePtr = readPtr;
      return readPtr;
    }

  }

}
