// Copyright (c) 2007 Paul Miner <$firstname.$lastname@gmail.com>
// This software is provided 'as-is', without any express or implied warranty. In no event will the authors be held liable for any damages arising from the use of this software.
// Permission is granted to anyone to use this software for any purpose, including commercial applications, and to alter it and redistribute it freely, subject to the following restrictions:
//     1. The origin of this software must not be misrepresented; you must not claim that you wrote the original software. If you use this software in a product, an acknowledgment in the product documentation would be appreciated but is not required.
//     2. Altered source versions must be plainly marked as such, and must not be misrepresented as being the original software.
//     3. This notice may not be removed or altered from any source distribution.

package org.blinkenbyte.io;

import java.io.EOFException;
import java.io.InputStream;
import java.io.IOException;

/**
 * This is a utility class containing methods to help deal with I/O operations.
 */
public class IOUtil
{

  private IOUtil()
  {  }
  
  /* ******************* InputStream ********************** */
  
  /**
   * Reads <i>b.length</i> bytes into <i>b</i> or throws an exception if end-of-stream is reached before all bytes can be read.
   */
  public static void readFully(InputStream inpStr, byte[] b) throws IOException
  {
    readFully(inpStr, b, 0, b.length);
    return;
  }
  
  /**
   * Reads <i>len</i> bytes into <i>b</i> at <i>b[offset]</i> or throws an exception if end-of-stream is reached before all bytes can be read.
   */
  public static void readFully(InputStream inpStr, byte[] b, int offset, int len) throws IOException
  {
    int bytesRead = fillBuffer(inpStr, b, offset, len);
    if (bytesRead != len)
      throw new EOFException("Only able to read " + bytesRead + " of " + len + " bytes before reaching EOF");
    return;
  }
  
  /**
   * Reads up to <i>b.length</i> bytes into <i>b</i>. Repeated reads are made until either <i>b.length</i> bytes are read or the end-of-stream
   * is reached.
   */
  public static int fillBuffer(InputStream inpStr, byte[] b) throws IOException
  {
    return fillBuffer(inpStr, b, 0, b.length);
  }
  
  /**
   * Reads up to <i>len</i> bytes into <i>b</i> at <i>b[offset]</i>. Repeated reads are made until either <i>len</i> bytes are read
   * or the end-of-stream is reached.
   */
  public static int fillBuffer(InputStream inpStr, byte[] b, int offset, int len) throws IOException
  {
    int totalRead = 0;
    while (totalRead < len)
    {
      int bytesRead = inpStr.read(b, offset, len);
      if (bytesRead <= 0)
        return totalRead;
      totalRead += bytesRead;
      offset += bytesRead;
      len -= bytesRead;
    }
    return totalRead;
  }
  
  /* ******************* ReadWriteDevice ********************** */

  /**
   * Reads <i>b.length</i> bytes into <i>b</i> or throws an exception if end-of-stream is reached before all bytes can be read.
   */
  public static void readFully(ReadWriteDevice rwd, byte[] b) throws IOException
  {
    readFully(rwd.getInputStream(), b, 0, b.length);
    return;
  }
  
  /**
   * Reads <i>len</i> bytes into <i>b</i> at <i>b[offset]</i> or throws an exception if end-of-stream is reached before all bytes can be read.
   */
  public static void readFully(ReadWriteDevice rwd, byte[] b, int offset, int len) throws IOException
  {
    readFully(rwd.getInputStream(), b, offset, len);
    return;
  }
  
  /**
   * Reads up to <i>b.length</i> bytes into <i>b</i>. Repeated reads are made until either <i>b.length</i> bytes are read or the end-of-stream
   * is reached.
   */
  public static int fillBuffer(ReadWriteDevice rwd, byte[] b) throws IOException
  {
    return fillBuffer(rwd.getInputStream(), b, 0, b.length);
  }
  
  /**
   * Reads up to <i>len</i> bytes into <i>b</i> at <i>b[offset]</i>. Repeated reads are made until either <i>len</i> bytes are read
   * or the end-of-stream is reached.
   */
  public static int fillBuffer(ReadWriteDevice rwd, byte[] b, int offset, int len) throws IOException
  {
    return fillBuffer(rwd.getInputStream(), b, offset, len);
  }

  /* ******************* BlockDevice ********************** */

  /**
   * Reads <i>b.length</i> bytes into <i>b</i> or throws an exception if end-of-stream is reached before all bytes can be read.
   */
  public static void readFully(BlockDevice bd, byte[] b) throws IOException
  {
    readFully(bd, b, 0, b.length);
    return;
  }
  
  /**
   * Reads <i>len</i> bytes into <i>b</i> at <i>b[offset]</i> or throws an exception if end-of-stream is reached before all bytes can be read.
   */
  public static void readFully(BlockDevice bd, byte[] b, int offset, int len) throws IOException
  {
    int bytesRead = fillBuffer(bd, b, offset, len);
    if (bytesRead != len)
      throw new EOFException("Only able to read " + bytesRead + " of " + len + " bytes before reaching EOF");
    return;
  }
  
  /**
   * Reads up to <i>b.length</i> bytes into <i>b</i>. Repeated reads are made until either <i>b.length</i> bytes are read or the end-of-stream
   * is reached.
   */
  public static int fillBuffer(BlockDevice bd, byte[] b) throws IOException
  {
    return fillBuffer(bd, b, 0, b.length);
  }
  
  /**
   * Reads up to <i>len</i> bytes into <i>b</i> at <i>b[offset]</i>. Repeated reads are made until either <i>len</i> bytes are read
   * or the end-of-stream is reached.
   */
  public static int fillBuffer(BlockDevice bd, byte[] b, int offset, int len) throws IOException
  {
    int totalRead = 0;
    while (totalRead < len)
    {
      int bytesRead = bd.read(b, offset, len);
      if (bytesRead <= 0)
        return totalRead;
      totalRead += bytesRead;
      offset += bytesRead;
      len -= bytesRead;
    }
    return totalRead;
  }
  

}
