// Copyright (c) 2007 Paul Miner <$firstname.$lastname@gmail.com>
// This software is provided 'as-is', without any express or implied warranty. In no event will the authors be held liable for any damages arising from the use of this software.
// Permission is granted to anyone to use this software for any purpose, including commercial applications, and to alter it and redistribute it freely, subject to the following restrictions:
//     1. The origin of this software must not be misrepresented; you must not claim that you wrote the original software. If you use this software in a product, an acknowledgment in the product documentation would be appreciated but is not required.
//     2. Altered source versions must be plainly marked as such, and must not be misrepresented as being the original software.
//     3. This notice may not be removed or altered from any source distribution.

package org.blinkenbyte.io;

import java.io.EOFException;
import java.io.IOException;

/**
 * This class implements a normal BlockDevice by wrapping a ReadWriteDevice.
 */
public class NormalBlockDevice implements BlockDevice
{

  private ReadWriteDevice rwd = null;
  private boolean sectorMode = false;
  private int sectorSize = 1;
  private int sectorSizeLog = 0;
  private boolean autoSectorAlign = false;

  /**
   * Constructs a new uninitialized NormalBlockDevice.
   */
  public NormalBlockDevice()
  {  }
  
  public void close() throws IOException
  {
    rwd.close();
    return;
  }
  
  /**
   * Sets the internal ReadWriteDevice this object wraps, and whether sector mode should be enforced.
   */
  public void setReadWriteDevice(ReadWriteDevice _rwd, boolean _sectorMode)
  {
    rwd = _rwd;
    sectorMode = _sectorMode;
    return;
  }
  
  /**
   * Uses the specified filename and parameters to create a new {@link org.blinkenbyte.io.ReadWriteFile ReadWriteFile} which will be wrapped by this object.
   */
  public void open(String filename, boolean allowWrite, boolean syncWriteData, boolean syncWriteMetaData, boolean _sectorMode) throws Exception
  {
    sectorMode = _sectorMode;
    String mode = null;
    if (allowWrite)
    {
      if (syncWriteData)
      {
        if (syncWriteMetaData)
          mode = "rws";
        else
          mode = "rwd";
      }
      else
        mode = "rw";
    }
    else
      mode = "r";
    rwd = new ReadWriteFile(filename, mode);
  }
  
  public long getFilePointer() throws IOException
  {
    return rwd.getPointer();
  }
  
  public long length() throws IOException
  {
    return rwd.length();
  }
  
  public int read() throws IOException
  {
    if (sectorMode && (sectorSizeLog > 0) && !autoSectorAlign)
      throw new IOException("Single-byte read requested not allowed in sector mode");
    return rwd.read();
  }
  
  public int read(byte[] b) throws IOException
  {
    return read(b, 0, b.length);
  }
  
  public int read(byte[] b, int offset, int len) throws IOException
  {
    if (sectorMode && (sectorSizeLog > 0))
    {
      long curPos = getFilePointer();
      int firstSectorOffset = (int)(curPos & (sectorSize - 1));
      boolean offsetIsAligned = (firstSectorOffset == 0);
      boolean isSizeAligned = (len & (sectorSize - 1)) == 0;
      if ((!offsetIsAligned || !isSizeAligned) && !autoSectorAlign)
        throw new IOException("Read size or offset is not aligned on sector boundaries");
    }
    return rwd.read(b, offset, len);
  }
  
  public void seek(long pos) throws IOException
  {
    rwd.seek(pos);
    return;
  }
  
  public int skipBytes(int n) throws IOException
  {
    return (int)rwd.skip((int)n);
  }
  
  public void write(byte[] b) throws IOException
  {
    rwd.write(b, 0, b.length);
    return;
  }
  
  public void write(byte[] b, int offset, int len) throws IOException
  {
    if (sectorMode && (sectorSizeLog > 0))
    {
      long curPos = getFilePointer();
      int firstSectorOffset = (int)(curPos & (sectorSize - 1));
      boolean offsetIsAligned = (firstSectorOffset == 0);
      boolean isSizeAligned = (len & (sectorSize - 1)) == 0;
      if ((!offsetIsAligned || !isSizeAligned) && !autoSectorAlign)
        throw new IOException("Write size or offset is not aligned on sector boundaries");
    }
    rwd.write(b, offset, len);
    return;
  }
  
  public void write(int b) throws IOException
  {
    if (sectorMode && (sectorSizeLog > 0) && !autoSectorAlign)
      throw new IOException("Single-byte write requested not allowed in sector mode");
    rwd.write(b);
    return;
  }

  /**
   * Throws an exception; since this class is normally used for files, it has no geometry.
   */
  public BlockDevice.DriveGeometry getDriveGeometry() throws Exception
  {
    throw new Exception("Operation unsupported");
  }
  
  public int getSectorSize()
  {
    return sectorSize;
  }
  
  public void setSectorSize(int _sectorSize) throws Exception
  {
    int _sectorSizeLog = 0;
    int n = _sectorSize;
    while ((n >>>= 1) != 0)
      _sectorSizeLog++;
    if (_sectorSize != (1 << _sectorSizeLog))
      throw new Exception("Invalid sector size (" + _sectorSize + " given, must be power of 2)");
    sectorSize = _sectorSize;
    sectorSizeLog = _sectorSizeLog;
    return;
  }
  
  public int getSectorSizeLog()
  {
    return sectorSizeLog;
  }

  public boolean getAutoSectorAlign()
  {
    return autoSectorAlign;
  }
  
  public void setAutoSectorAlign(boolean _autoSectorAlign)
  {
    autoSectorAlign = _autoSectorAlign;
    return;
  }

}
