// Copyright (c) 2007 Paul Miner <$firstname.$lastname@gmail.com>
// This software is provided 'as-is', without any express or implied warranty. In no event will the authors be held liable for any damages arising from the use of this software.
// Permission is granted to anyone to use this software for any purpose, including commercial applications, and to alter it and redistribute it freely, subject to the following restrictions:
//     1. The origin of this software must not be misrepresented; you must not claim that you wrote the original software. If you use this software in a product, an acknowledgment in the product documentation would be appreciated but is not required.
//     2. Altered source versions must be plainly marked as such, and must not be misrepresented as being the original software.
//     3. This notice may not be removed or altered from any source distribution.

package org.blinkenbyte.io;

import java.io.InputStream;
import java.io.IOException;
import java.io.OutputStream;

/**
 * This class implements a ReadWriteDevice using a ByteArrayBuffer as a backing store.
 */
public class ReadWriteByteArray implements ReadWriteDevice
{
  protected ByteArrayBuffer bab = null;
  protected ByteArrayBuffer.ByteArrayBufferInputStream babInp = null;
  protected boolean isClosed = false;
  
  public ReadWriteByteArray()
  {
    this(0);
  }
  
  /**
   * Constructs a new ReadWriteByteArray with an initial size of <i>initialSize</i> bytes.
   */
  public ReadWriteByteArray(int initialSize)
  {
    bab = new ByteArrayBuffer();
    babInp = bab.getInputStream(true);
    bab.setLength(initialSize);
  }
  
  public void close() throws IOException
  {
    throwIfClosed();
    isClosed = true;
    babInp.close();
    babInp = null;
    bab.close();
    bab = null;
  }
  
  protected void throwIfClosed() throws IOException
  {
    if (isClosed)
      throw new IOException("ReadWriteByteArray is already closed");
    return;
  }

  public void flush() throws IOException
  {
    throwIfClosed();
    return;
  }
  
  public int available() throws IOException
  {
    throwIfClosed();
    return bab.length() - bab.getWritePtr();
  }
  
  public long length() throws IOException
  {
    throwIfClosed();
    return bab.length();
  }
  
  public void setLength(int newLength) throws IOException
  {
    throwIfClosed();
    bab.setLength(newLength);
    return;
  }
  
  public long getPointer() throws IOException
  {
    throwIfClosed();
    return bab.getWritePtr();
  }
  
  public void seek(long newPointer) throws IOException
  {
    throwIfClosed();
    bab.seek((int)newPointer);
    return;
  }
  
  public long skip(long amtToSkip) throws IOException
  {
    throwIfClosed();
    return babInp.skip(amtToSkip);
  }
  
  public int read() throws IOException
  {
    throwIfClosed();
    return babInp.read();
  }
  
  public int read(byte[] b) throws IOException
  {
    throwIfClosed();
    return babInp.read(b, 0, b.length);
  }
  
  public int read(byte[] b, int offset, int len) throws IOException
  {
    throwIfClosed();
    return babInp.read(b, offset, len);
  }

  public void write(int b) throws IOException
  {
    throwIfClosed();
    bab.write(b);
    return;
  }
  
  public void write(byte[] b) throws IOException
  {
    throwIfClosed();
    bab.write(b, 0, b.length);
    return;
  }
  
  public void write(byte[] b, int offset, int len) throws IOException
  {
    throwIfClosed();
    bab.write(b, offset, len);
    return;
  }

  public void mark(long markLimit) throws IOException
  {
    throwIfClosed();
    babInp.mark((int)markLimit);
    return;
  }
  
  public boolean markSupported()
  {
    return babInp.markSupported();
  }
  
  public void reset() throws IOException
  {
    throwIfClosed();
    babInp.reset();
    return;
  }
  
  public void clear()
  {
    bab.setLength(0);
    bab.fillZeroes();
    return;
  }

  public InputStream getInputStream()
  {
    return babInp;
  }
  
  public OutputStream getOutputStream()
  {
    return bab;
  }

}
