// Copyright (c) 2007 Paul Miner <$firstname.$lastname@gmail.com>
// This software is provided 'as-is', without any express or implied warranty. In no event will the authors be held liable for any damages arising from the use of this software.
// Permission is granted to anyone to use this software for any purpose, including commercial applications, and to alter it and redistribute it freely, subject to the following restrictions:
//     1. The origin of this software must not be misrepresented; you must not claim that you wrote the original software. If you use this software in a product, an acknowledgment in the product documentation would be appreciated but is not required.
//     2. Altered source versions must be plainly marked as such, and must not be misrepresented as being the original software.
//     3. This notice may not be removed or altered from any source distribution.

package org.blinkenbyte.io;

import java.io.Closeable;
import java.io.InputStream;
import java.io.IOException;
import java.io.OutputStream;

/**
 * Interface for defining stream oriented read/write devices.
 */
public interface ReadWriteDevice extends Closeable
{

  /**
   * Closes the device.
   */
  public void close() throws IOException;
  
  /**
   * Flushes the device.
   */
  public void flush() throws IOException;

  /**
   * Returns the number of bytes between the current pointer and end-of-stream.
   */
  public int available() throws IOException;
  
  /**
   * Returns the total length in bytes of this device.
   */
  public long length() throws IOException;
  
  /**
   * Returns the current pointer for this device.
   */
  public long getPointer() throws IOException;
  
  /**
   * Seeks the pointer to the specified offset.
   */
  public void seek(long newPointer) throws IOException;

  /**
   * Advances the pointer by the specified number of bytes
   */
  public long skip(long amtToSkip) throws IOException;
  
  /**
   * Reads and returns a single byte. Returns -1 if end-of-stream.
   */
  public int read() throws IOException;

  /**
   * Reads up to <i>b.length</i> bytes from the device into array <i>b</i>. Returns the number of bytes read,
   * -1 if currently at end-of-stream.
   */
  public int read(byte[] b) throws IOException;

  /**
   * Reads up to <i>len</i> bytes from the device into array <i>b</i> at <i>b[offset]</i>. Returns the number of bytes read,
   * -1 if currently at end-of-stream.
   */
  public int read(byte[] b, int offset, int len) throws IOException;

  /**
   * Writes a single byte to the device.
   */
  public void write(int b) throws IOException;

  /**
   * Writes <i>b.length</i> bytes from array <i>b</i> to the device.
   */
  public void write(byte[] b) throws IOException;

  /**
   * Writes <i>len</i> bytes from array <i>b</i> starting at <i>b[offset]</i> to the device.
   */
  public void write(byte[] b, int offset, int len) throws IOException;

  /**
   * Sets the mark with the specified limit.
   */
  public void mark(long markLimit) throws IOException;

  /**
   * Returns whether mark is supported.
   */
  public boolean markSupported();

  /**
   * Resets the pointer to the mark position.
   */
  public void reset() throws IOException;

  /**
   * Return an InputStream for reading from this device.
   */
  public InputStream getInputStream();

  /**
   * Return an OututStream for writing to this device.
   */
  public OutputStream getOutputStream();

  /**
   * Implementation of an InputStream wrapped around a ReadWriteDevice.
   */
  public static class RWDInputStream extends InputStream
  {
    /**
     * The ReadWriteDevice being wrapped.
     */
    public ReadWriteDevice rwd = null;

    /**
     * Flag indicating if this stream has been closed.
     */
    protected boolean isClosed = false;
    
    /**
     * Constructs a new RWDInputStream from a ReadWriteDevice.
     */
    public RWDInputStream(ReadWriteDevice _rwd)
    {
      rwd = _rwd;
    }
    
    /**
     * Convenience method to throw an exception if the stream has been closed.
     */
    protected void throwIfClosed() throws IOException
    {
      if (isClosed)
        throw new IOException("RWDInputStream is already closed");
      return;
    }
    
    /**
     * Returns the number of bytes available to be read without blocking.
     */
    public int available() throws IOException
    {
      throwIfClosed();
      return rwd.available();
    }
    
    /**
     * Closes this stream (but not the underlying ReadWriteDevice).
     */
    public void close() throws IOException
    {
      throwIfClosed();
      isClosed = true;
      rwd = null;
      return;
    }
    
    /**
     * Marks the input stream.
     */
    public void mark(int readLimit)
    {
      try
      {  rwd.mark(readLimit);  }
      catch (Exception e)
      {  }
      return;
    }
    
    /**
     * Returns whether the ReadWriteDevice being wrapped supports mark.
     */
    public boolean markSupported()
    {
      return rwd.markSupported();
    }
    
    /**
     * Reads a byte from the underlying ReadWriteDevice. Returns -1 if end-of-stream.
     */
    public int read() throws IOException
    {
      throwIfClosed();
      return rwd.read();
    }
    
    /**
     * Reads up to <i>b.length</i> bytes from the device into array <i>b</i>. Returns the number of bytes read,
     * -1 if currently at end-of-stream.
     */
    public int read(byte[] b) throws IOException
    {
      throwIfClosed();
      return rwd.read(b, 0, b.length);
    }
    
    /**
     * Reads up to <i>len</i> bytes from the device into array <i>b</i> at <i>b[offset]</i>. Returns the number of bytes read,
     * -1 if currently at end-of-stream.
     */
    public int read(byte[] b, int offset, int len) throws IOException
    {
      throwIfClosed();
      return rwd.read(b, offset, len);
    }

    /**
     * Resets the stream to the mark.
     */    
    public void reset() throws IOException
    {
      throwIfClosed();
      rwd.reset();
      return;
    }
    
    /**
     * Skips <i>n</i> bytes from the stream.
     */
    public long skip(long n) throws IOException
    {
      throwIfClosed();
      return rwd.skip(n);
    }
  }
  
  /**
   * Implementation of an OutputStream wrapped around a ReadWriteDevice.
   */
  public static class RWDOutputStream extends OutputStream
  {
    /**
     * The ReadWriteDevice being wrapped.
     */
    public ReadWriteDevice rwd = null;

    /**
     * Flag indicating if this stream has been closed.
     */
    protected boolean isClosed = false;
    
    /**
     * Constructs a new RWDOutputStream from a ReadWriteDevice.
     */
    public RWDOutputStream(ReadWriteDevice _rwd)
    {
      rwd = _rwd;
    }
    
    /**
     * Convenience method to throw an exception if the stream has been closed.
     */
    protected void throwIfClosed() throws IOException
    {
      if (isClosed)
        throw new IOException("RWDInputStream is already closed");
      return;
    }

    /**
     * Closes this stream (but not the underlying ReadWriteDevice).
     */
    public void close() throws IOException
    {
      throwIfClosed();
      isClosed = true;
      rwd = null;
      return;
    }
    
    /**
     * Flushes this stream.
     */
    public void flush() throws IOException
    {
      throwIfClosed();
      rwd.flush();
      return;
    }
    
    /**
     * Writes a single byte to the stream.
     */
    public void write(int b) throws IOException
    {
      throwIfClosed();
      rwd.write(b);
      return;
    }
    
    /**
     * Writes <i>b.length</i> bytes from array <i>b</i> to the stream.
     */
    public void write(byte[] b) throws IOException
    {
      throwIfClosed();
      rwd.write(b, 0, b.length);
      return;
    }
    
    /**
     * Writes <i>len</i> bytes from array <i>b</i> at offset <i>b[offset]</i> to the stream.
     */
    public void write(byte[] b, int offset, int len) throws IOException
    {
      throwIfClosed();
      rwd.write(b, offset, len);
      return;
    }
    
  }
}
