// Copyright (c) 2007 Paul Miner <$firstname.$lastname@gmail.com>
// This software is provided 'as-is', without any express or implied warranty. In no event will the authors be held liable for any damages arising from the use of this software.
// Permission is granted to anyone to use this software for any purpose, including commercial applications, and to alter it and redistribute it freely, subject to the following restrictions:
//     1. The origin of this software must not be misrepresented; you must not claim that you wrote the original software. If you use this software in a product, an acknowledgment in the product documentation would be appreciated but is not required.
//     2. Altered source versions must be plainly marked as such, and must not be misrepresented as being the original software.
//     3. This notice may not be removed or altered from any source distribution.

package org.blinkenbyte.io;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.InputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.io.RandomAccessFile;
import org.blinkenbyte.io.ReadWriteDevice.RWDInputStream;
import org.blinkenbyte.io.ReadWriteDevice.RWDOutputStream;
/**
 * Class to present a file as a ReadWriteDevice. Essentially an extension of RandomAccessFile
 */
public class ReadWriteFile implements ReadWriteDevice
{

  /**
   * The RandomAccessFile being wrapped
   */
  protected RandomAccessFile raf = null;
  
  /**
   * The current mark.
   */
  protected long markPtr = -1L;
  
  /**
   * Flag indicating whether this ReadWriteFile has been closed.
   */
  protected boolean isClosed = false;

  /**
   * Constructs a new ReadWriteFile around the specified RandomAccessFile
   */
  public ReadWriteFile(RandomAccessFile _raf) throws IOException
  {
    raf = _raf;
  }
  
  /**
   * Constructs a new ReadWriteFile using the specified File and mode. The mode is the same format as used in
   * {@link java.io.RandomAccessFile RandomAccessFile}.
   */
  public ReadWriteFile(File file, String mode) throws FileNotFoundException
  {
    raf = new RandomAccessFile(file, mode);
  }
  
  /**
   * Constructs a new ReadWriteFile using the specified filename and mode. The mode is the same format as used in
   * {@link java.io.RandomAccessFile RandomAccessFile}.
   */
  public ReadWriteFile(String filename, String mode) throws FileNotFoundException
  {
    raf = new RandomAccessFile(filename, mode);
  }
  
  /**
   * Returns 0.
   */
  public int available() throws IOException
  {
    return 0;
  }
  
  /**
   * Closes this object and the underlying RandomAccessFile.
   */
  public void close() throws IOException
  {
    raf.close();
    return;
  }
  
  /**
   * Flushes the underlying RandomAccessFile.
   */
  public void flush() throws IOException
  {
    raf.getChannel().force(true);
    return;
  }
  
  /**
   * Convenience method to throw an exception if this object has already been closed.
   */
  protected void throwIfClosed() throws IOException
  {
    if (isClosed)
      throw new IOException("File has already been closed");
    return;
  }
  
  /**
   * Returns the length of the underlying file.
   */
  public long length() throws IOException
  {
    return raf.length();
  }
  
  /**
   * Returns the current file pointer.
   */
  public long getPointer() throws IOException
  {
    return raf.getFilePointer();
  }
  
  /**
   * Seeks the underlying file to the specified position.
   */
  public void seek(long newPointer) throws IOException
  {
    if (newPointer < 0)
      raf.seek(raf.length() + newPointer);
    else
      raf.seek(newPointer);
    return;
  }
  
  /**
   * Advances the file pointer by the specified number of bytes.
   */
  public long skip(long n) throws IOException
  {
    long oldPos = raf.getFilePointer();
    raf.seek(oldPos + n);
    return raf.getFilePointer() - oldPos;
  }
  
  /**
   * Reads a byte from the underlying ReadWriteDevice. Returns -1 if end-of-stream.
   */
  public int read() throws IOException
  {
    return raf.read();
  }
  
  /**
   * Reads up to <i>b.length</i> bytes from the device into array <i>b</i>. Returns the number of bytes read,
   * -1 if currently at end-of-stream.
   */
  public int read(byte[] b) throws IOException
  {
    return raf.read(b, 0, b.length);
  }
  
  /**
   * Reads up to <i>len</i> bytes from the device into array <i>b</i> at <i>b[offset]</i>. Returns the number of bytes read,
   * -1 if currently at end-of-stream.
   */
  public int read(byte[] b, int offset, int len) throws IOException
  {
    return raf.read(b, offset, len);
  }
  
  /**
   * Writes a single byte to the stream.
   */
  public void write(int b) throws IOException
  {
    raf.write(b);
    return;
  }
  
  /**
   * Writes <i>b.length</i> bytes from array <i>b</i> to the stream.
   */
  public void write(byte[] b) throws IOException
  {
    raf.write(b, 0, b.length);
    return;
  }
  
  /**
   * Writes <i>len</i> bytes from array <i>b</i> at offset <i>b[offset]</i> to the stream.
   */
  public void write(byte[] b, int offset, int len) throws IOException
  {
    raf.write(b, offset, len);
    return;
  }

  /**
   * Sets the mark with the specified mark limit.
   */
  public void mark(long markLimit) throws IOException
  {
    markPtr = getPointer();
    return;
  }
  
  /**
   * Returns true.
   */
  public boolean markSupported()
  {
    return true;
  }
  
  /**
   * Resets the pointer to the last mark.
   */
  public void reset() throws IOException
  {
    if (markPtr < 0L)
      throw new IOException("No mark");
    seek(markPtr);
    return;
  }

  /**
   * Returns a new RWDInputStream wrapped around this object.
   */
  public RWDInputStream getInputStream()
  {
    return new RWDInputStream(this);
  }
  
  /**
   * Returns a new RWDOutputStream wrapped around this object.
   */
  public RWDOutputStream getOutputStream()
  {
    return new RWDOutputStream(this);
  }
  
}
