// Copyright (c) 2007 Paul Miner <$firstname.$lastname@gmail.com>
// This software is provided 'as-is', without any express or implied warranty. In no event will the authors be held liable for any damages arising from the use of this software.
// Permission is granted to anyone to use this software for any purpose, including commercial applications, and to alter it and redistribute it freely, subject to the following restrictions:
//     1. The origin of this software must not be misrepresented; you must not claim that you wrote the original software. If you use this software in a product, an acknowledgment in the product documentation would be appreciated but is not required.
//     2. Altered source versions must be plainly marked as such, and must not be misrepresented as being the original software.
//     3. This notice may not be removed or altered from any source distribution.

package org.blinkenbyte.io;

import java.io.EOFException;
import java.io.IOException;

/**
 * This class uses JNI to access block devices under Windows. Typically, the filename must begin with \\?\ for devices.
 * An example would be "\\?\PhysicalDrive0". This class requires WindowsBlockDevice.dll
 */
public class WindowsBlockDevice implements BlockDevice
{

  private byte[] fileHandle = null;
  private boolean sectorMode = false;
  private int sectorSize = 0;
  private int sectorSizeLog = 0;
  private boolean autoSectorAlign = false;

  static
  {
    System.loadLibrary("WindowsBlockDevice");
  }
  
  public WindowsBlockDevice()
  {  }
  
  public void close() throws IOException
  {
    throwIfNotOpen();
    nativeClose(fileHandle);
    return;
  }

  public void open(String filename, boolean allowWrite, boolean syncWriteData, boolean syncWriteMetaData, boolean _sectorMode) throws Exception
  {
    fileHandle = nativeOpen(filename, allowWrite, syncWriteData, syncWriteMetaData, _sectorMode);
    sectorMode = _sectorMode;
    if (sectorMode && (sectorSize == 0))
    {
      BlockDevice.DriveGeometry geometry = getDriveGeometry();
      if (geometry == null)
        throw new Exception("Unable to obtain drive geometry and no sector size specified");
      setSectorSize(geometry.numBytesPerSector);
    }
    return;
  }
  
  public void throwIfNotOpen() throws IOException
  {
    if (fileHandle == null)
      throw new IOException("File not open");
    return;
  }
  
  public long getFilePointer() throws IOException
  {
    throwIfNotOpen();
    return nativeGetFilePointer(fileHandle);
  }

  public long length() throws IOException
  {
    throwIfNotOpen();
    long len = -1L;
    try
    {
      return nativeLength(fileHandle);
    }
    catch (IOException ioe)
    {
      try
      {
        BlockDevice.DriveGeometry geometry = getDriveGeometry();
        if (geometry == null)
          return -1L;
        return geometry.diskSize;
      }
      catch (Exception ioe2)
      {  }
      throw ioe;
    }
  }

  public int read() throws IOException
  {
    throwIfNotOpen();
    if (sectorMode && (sectorSizeLog > 0))
    {
      if (!autoSectorAlign)
        throw new IOException("Single-byte read requested not allowed in sector mode");
      long curPos = getFilePointer();
      byte[] buffer = new byte[sectorSize];
      nativeSeek(fileHandle, curPos & ~((long)(sectorSize - 1)));
      int bytesRead = read(buffer);
      if (bytesRead != sectorSize)
      {
        if (bytesRead <= 0)
          return -1;
        throw new IOException("Unable to read sector");
      }
      int returnByte = (int)buffer[(int)(curPos++ & (sectorSize - 1))] & 0xFF;
      nativeSeek(fileHandle, curPos);
      return returnByte;
    }
    return nativeRead(fileHandle);
  }

  public int read(byte[] b) throws IOException
  {
    return read(b, 0, b.length);
  }
  
  public int read(byte[] b, int offset, int len) throws IOException
  {
    throwIfNotOpen();
    if (sectorMode && (sectorSizeLog > 0))
    {
      long curPos = getFilePointer();
      int firstSectorOffset = (int)(curPos & (sectorSize - 1));
      boolean offsetIsAligned = (firstSectorOffset == 0);
      boolean isSizeAligned = (len & (sectorSize - 1)) == 0;
      
      if (!offsetIsAligned || !isSizeAligned)
      {
        if (!autoSectorAlign)
          throw new IOException("Read size or offset is not aligned on sector boundaries");
        byte[] buffer = new byte[sectorSize];
        int originalLen = len;
        if (firstSectorOffset > 0)
        {
          nativeSeek(fileHandle, curPos & ~((int)(sectorSize - 1)));
          int bytesRead = nativeReadArray(fileHandle, buffer, 0, sectorSize, sectorSizeLog);
          if (bytesRead != sectorSize)
          {
            if (bytesRead <= 0)
              return -1;
            throw new IOException("Unable to read full sector");
          }
          bytesRead = sectorSize - firstSectorOffset;
          boolean readExtra = (bytesRead > len);
          if (readExtra)
            bytesRead = len;
          System.arraycopy(buffer, firstSectorOffset, b, offset, bytesRead);
          len -= bytesRead;
          offset += bytesRead;
          if (readExtra)
            nativeSeek(fileHandle, curPos + originalLen);
        }
        int numFullSectors = len >>> sectorSizeLog;
        if (numFullSectors > 0)
        {
          int requestedLen = numFullSectors << sectorSizeLog;
          int bytesRead = nativeReadArray(fileHandle, b, offset, requestedLen, sectorSizeLog);
          if (bytesRead <= 0)
          {
            if (len != originalLen)
              return (originalLen - len);
            return -1;
          }
          len -= bytesRead;
          offset += bytesRead;
          if (bytesRead != requestedLen)
            return (originalLen - len);
        }
        if (len > 0)
        {
          int bytesRead = nativeReadArray(fileHandle, buffer, 0, sectorSize, sectorSizeLog);
          if (bytesRead <= 0)
          {
            if (len != originalLen)
              return (originalLen - len);
            return -1;
          }
          if (bytesRead != sectorSize)
            throw new IOException("Unable to read full sector");
          System.arraycopy(buffer, 0, b, offset, len);
          offset += len;
          len = 0;
          nativeSeek(fileHandle, curPos + originalLen);
        }
        return (originalLen - len);
      }
    }
    return nativeReadArray(fileHandle, b, offset, len, sectorMode ? sectorSizeLog : 0);
  }
  
  public void seek(long pos) throws IOException
  {
    throwIfNotOpen();
    nativeSeek(fileHandle, pos);
  }
  
  public int skipBytes(int n) throws IOException
  {
    throwIfNotOpen();
    return nativeSkipBytes(fileHandle, n);
  }
  
  public void write(byte[] b) throws IOException
  {
    write(b, 0, b.length);
    return;
  }
  
  public void write(byte[] b, int offset, int len) throws IOException
  {
    throwIfNotOpen();
    if (sectorMode && (sectorSizeLog > 0))
    {
      long curPos = getFilePointer();
      int firstSectorOffset = (int)(curPos & (sectorSize - 1));
      boolean offsetIsAligned = (firstSectorOffset == 0);
      boolean isSizeAligned = (len & (sectorSize - 1)) == 0;
      
      if (!offsetIsAligned || !isSizeAligned)
      {
        if (!autoSectorAlign)
          throw new IOException("Write size or offset is not aligned on sector boundaries");
        byte[] buffer = new byte[sectorSize];
        int originalLen = len;
        if (firstSectorOffset > 0)
        {
          nativeSeek(fileHandle, curPos & ~((int)(sectorSize - 1)));
          int bytesRead = nativeReadArray(fileHandle, buffer, 0, sectorSize, sectorSizeLog);
          if (bytesRead != sectorSize)
            throw new IOException("Unable to read sector to update");
          bytesRead = sectorSize - firstSectorOffset;
          boolean readExtra = (bytesRead > len);
          if (readExtra)
            bytesRead = len;
          System.arraycopy(b, offset, buffer, firstSectorOffset, bytesRead);
          nativeSeek(fileHandle, curPos & ~((int)(sectorSize - 1)));
          nativeWriteArray(fileHandle, buffer, 0, sectorSize, sectorSizeLog);
          len -= bytesRead;
          offset += bytesRead;
          if (readExtra)
            nativeSeek(fileHandle, curPos + originalLen);
        }
        int numFullSectors = len >>> sectorSizeLog;
        if (numFullSectors > 0)
        {
          int requestedLen = numFullSectors << sectorSizeLog;
          write(b, offset, requestedLen);
          len -= requestedLen;
          offset += requestedLen;
        }
        if (len > 0)
        {
          int bytesRead = nativeReadArray(fileHandle, buffer, 0, sectorSize, sectorSizeLog);
          if (bytesRead != sectorSize)
            throw new IOException("Unable to read sector to update");
          System.arraycopy(b, offset, buffer, 0, len);
          nativeSeek(fileHandle, curPos + originalLen - len);
          nativeWriteArray(fileHandle, buffer, 0, sectorSize, sectorSizeLog);
          offset += len;
          len = 0;
          nativeSeek(fileHandle, curPos + originalLen);
        }
        return;
      }
    }
    nativeWriteArray(fileHandle, b, offset, len, sectorMode ? sectorSizeLog : 0);
    return;
  }
  
  public void write(int b) throws IOException
  {
    throwIfNotOpen();
    if (sectorMode && (sectorSizeLog > 0))
    {
      if (!autoSectorAlign)
        throw new IOException("Single-byte read requested not allowed in sector mode");
      long curPos = getFilePointer();
      byte[] buffer = new byte[sectorSize];
      long basePos = curPos & ~((long)(sectorSize - 1));
      nativeSeek(fileHandle, basePos);
      int bytesRead = read(buffer);
      if (bytesRead != sectorSize)
        throw new IOException("Unable to read sector to update");
      buffer[(int)(curPos & (sectorSize - 1))] = (byte)b;
      nativeSeek(fileHandle, basePos);
      write(buffer);
      nativeSeek(fileHandle, curPos + 1);
      return;
    }
    nativeWrite(fileHandle, b);
    return;
  }
  
  public BlockDevice.DriveGeometry getDriveGeometry() throws Exception
  {
    throwIfNotOpen();
    return nativeGetDriveGeometry(fileHandle);
  }

  public int getSectorSize()
  {
    return sectorSize;
  }
  
  public void setSectorSize(int _sectorSize) throws Exception
  {
    int _sectorSizeLog = 0;
    int n = _sectorSize;
    while ((n >>>= 1) != 0)
      _sectorSizeLog++;
    if (_sectorSize != (1 << _sectorSizeLog))
      throw new Exception("Invalid sector size (" + _sectorSize + " given, must be power of 2)");
    sectorSize = _sectorSize;
    sectorSizeLog = _sectorSizeLog;
    return;
  }
  
  public int getSectorSizeLog()
  {
    return sectorSizeLog;
  }

  public boolean getAutoSectorAlign()
  {
    return autoSectorAlign;
  }
  
  public void setAutoSectorAlign(boolean _autoSectorAlign)
  {
    autoSectorAlign = _autoSectorAlign;
    return;
  }

  private static native void nativeClose(byte[] handle) throws IOException;
  private static native byte[] nativeOpen(String filename, boolean allowWrite, boolean syncWriteData, boolean syncWriteMetaData, boolean _sectorMode) throws Exception;
  private static native long nativeGetFilePointer(byte[] handle) throws IOException;
  private static native long nativeLength(byte[] handle) throws IOException;
  private static native int nativeRead(byte[] handle) throws IOException;
  private static native int nativeReadArray(byte[] handle, byte[] b, int offset, int len, int sectorSizeLog) throws IOException;
  private static native long nativeSeek(byte[] handle, long pos) throws IOException;
  private static native int nativeSkipBytes(byte[] handle, int n) throws IOException;
  private static native void nativeWriteArray(byte[] handle, byte[] b, int offset, int len, int sectorSizeLog) throws IOException;
  private static native void nativeWrite(byte[] handle, int b) throws IOException;
  private static native BlockDevice.DriveGeometry nativeGetDriveGeometry(byte[] handle) throws Exception;

}
