// Copyright (c) 2007 Paul Miner <$firstname.$lastname@gmail.com>
// This software is provided 'as-is', without any express or implied warranty. In no event will the authors be held liable for any damages arising from the use of this software.
// Permission is granted to anyone to use this software for any purpose, including commercial applications, and to alter it and redistribute it freely, subject to the following restrictions:
//     1. The origin of this software must not be misrepresented; you must not claim that you wrote the original software. If you use this software in a product, an acknowledgment in the product documentation would be appreciated but is not required.
//     2. Altered source versions must be plainly marked as such, and must not be misrepresented as being the original software.
//     3. This notice may not be removed or altered from any source distribution.

package org.blinkenbyte.pcopier;

import java.io.BufferedWriter;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.RandomAccessFile;
import java.io.Reader;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.NoSuchElementException;
import org.blinkenbyte.io.BadBlockEmulator;
import org.blinkenbyte.io.BlockDevice;
import org.blinkenbyte.io.BlockRange;
import org.blinkenbyte.io.IOUtil;
import org.blinkenbyte.io.NormalBlockDevice;
import org.blinkenbyte.io.ReadWriteByteArray;
import org.blinkenbyte.io.ReadWriteDevice;
import org.blinkenbyte.io.ReadWriteFile;

/**
 * The ProgressiveCopier is a utility to aid in imaging drives that have errors on them. A common problem with damaged drives is
 * that copying them may take an extraordinarily long time from start to finish, sector by sector. It may never finish, and with
 * damaged drives, getting as much data off as quickly as possible is important. Once the image has been created, it can be mounted
 * or opened directly by a data recovery program without having to worry about bad sectors preventing the program from scanning the drive.
 * <p>
 * This utility aids this process in two ways. First, it maintains a map of sectors and whether they have been read or not. Second
 * (and more importantly), it has a hierarchial strategy to copy the block device. The drive is divided up into blocks, and when copying
 * a block of data, any errors will cause it to mark the block as incomplete and move to the next block. Each of these blocks is in turn
 * divided into smaller blocks, forming a large hierarchy of blocks. The topmost (first) level 0 is a block large enough to encompass the
 * entire device. The bottom (last) level is composed of blocks the size of the input's sector size. The number of sub-blocks each block
 * is broken into defaults to 4, but can specified.
 * <p>
 * The advantage of this approach is that larger contiguous blocks of error-free data can be ready quickly, and the areas with errors are
 * left for later. As the copy progresses, the detail with which it attempts to read the bad areas is reduced to smaller and smaller blocks.
 * <p>
 * As an example, imagine starting with a 128GB hard drive. The utility will start reading the whole 128GB drive as a single large block.
 * If an error occurs 79GB into the drive, the 128GB block is marked as partial, and since it's the only block in the top level, the utility
 * progresses to the next level. At this level, the drive is divided into 32GB blocks. The first two blocks are already complete and the
 * third block is already partially read, so they are skipped. It then resumes at the fourth block. Suppose it completes fully, and moves
 * onto the next level. At this level, the blocks are 8GB in size, and the only region that needs to be dealt with is from 64-96GB. The block
 * from 64-72GB is complete, and the block from 72-79GB is marked as partial, so they will not be read. The blocks from 80-88GB and 88-96GB
 * are marked as unread, so they are read. After reading, it then continues to the next level, etc, until execution is stopped or the last
 * level is completed.
 * <p>
 * In cases where this utility is useful, the sector level will generally not be reached. If it is reached, it indicates that a straight
 * start-to-end sector by sector copy would have been faster. The only utility in that case is where the ability to resume a copy is needed.
 * <p>
 * The map file has a header followed by the map data, which uses 2 bits per block. The number of blocks will depend on the number of
 * sub-blocks (childrenPerNode) each block is divided into, and the virtual size of the input device. The virtual size of the input device
 * is smallest block size that is equal or larger in size than the input device. It can be determined by starting with the sector size
 * and repeatedly multiplying by childrenPerNode until a number greater or equal to the input device size is reached. For example, if the
 * sector size is 512 bytes, childrenPerNode is 4, and the device's size is  48K, the virtual size will be 512 * 4 * 4 * 4 * 4 = 128K.
 * This also tells me the number of levels and their block sizes. In this case, the levels are numbered 0-4. Level 0 is a single 128K
 * block, level 1 is four 32K blocks, level 2 is sixteen  8K blocks, level 3 is 64 2K blocks, and level 4 is 256 512-byte blocks (256
 * sectors). The status for the virtual blocks past the end of the input is maintained as if they were real (and always successful),
 * but read requests for these virtual blocks will never be passed to the input device.
 * <p>
 * The map file stores the status of every block in a large array of bits. The status of child blocks are not always updated; if a parent
 * block is marked as complete, the status of the children is irrelevant. This saves a considerable amount of I/O to the map file. The
 * storage is byte-oriented, so endian-ness is not a concern when reading the map. The low order bits of a byte are considered to come
 * before the high-order bits of a byte. Because each status is 2 bits, a single byte can hold four statuses. At the last level, one
 * byte of the map is responsible for 2048 bytes of input (assuming 512 byte sectors). Depending on childrenPerNode, the ratio between the
 * virtual size of the input device and the size of the map will be roughly in the range of 1024-2048 to 1. If childrenPerNode is 4, the ratio
 * is roughly 1536:1; assuming a worst-case scenario of the virtual size being 4 times as large as the actual input size, the ratio works out to
 * 384:1, or roughly 0.26% of the input size, so the overhead compared to the resulting output is minimal.
 * <p>
 * As an example of how the map file stores statuses, bits 1-0 of byte 0 of the map contain the status for level 0 block 0, bits 3-2 contain
 * the status for level 1 block 0, bits 5-4 for level 1 block 1, bits 7-6 for level 1 block 2, bits 1-0 of byte 1 for level 1 block 3,
 * bits 3-2 of byte 1 for level 2 block 0, etc.
 * <p>
 * To determine the status of a block, all of its parents must be checked. If a parent has a completed status, then all sectors that fall
 * within that block are considered complete regardless of their status. Note that complete does not necessarily mean successfully read,
 * just that every sector in the block has either been read or attempted to be read individually. A completed or partial block with a parent
 * that has an empty status is the result of a corrupt map and should never happen.
 * <p>
 * If a specific set of sectors should be skipped, a bad sector list can be used. A bad block emulator device will then be wrapped around the
 * input to simulate errors if those sectors are read from, and the read request will not be passed to the input device. Overlapping ranges
 * are allowed, and reads spanning good and bad areas are errored out or truncated as needed so that the read of the bad area is never seen
 * by the input device.
 * <p>
 * The format of a sector list is a series of entries on separate lines. Each line contains the starting sector number, and an optional sector
 * count. If the sector count is not specified, the default is 1. The numbers can be in either decimal or hexadecimal. Hexadecimal numbers
 * should be prefixed with "0x", as in "0x64" would be sector 100. Spaces, commas, and tabs can be used as the separator between the two values.
 * Both the input bad sector list and the output incomplete sector list use this format.
 * <p>
 * Basic operation: Specify an input, output, and map file or device, then call open(), performCopy(), and close().
 */
public class ProgressiveCopier
{

  /**
   * Status indicating an unread block.
   */
  public static final int STATUS_UNREAD = 0;

  /**
   * Status indicating a block that returned an error (only valid at the sector level).
   */
  public static final int STATUS_ERROR = 1;

  /**
   * Status indicating a block that has only been read partially either due to an error in a bulk transfer or an interrupted or ongoing transfer.
   */
  public static final int STATUS_PARTIAL = 1;

  /**
   * Status indicating a block that has been completed (all sectors are either read or unreadable).
   */
  public static final int STATUS_COMPLETE = 3;

  /**
   * Current map file version number
   */
  public static final int CURRENT_MAP_VERSION = 1;
  private static final byte[] MAPFILE_HEADER_MARKER = "PCMAP".getBytes();

  /**
   * Map file header size
   */
  public static final int MAPFILE_HEADER_SIZE = MAPFILE_HEADER_MARKER.length + 12;
  // Offset   Content
  //  0       String "PCMAP"
  //  5       DWORD little-endian versionNumber
  //  9       DWORD little-endian numChildrenPerNode
  // 13       DWORD little-endian maxLevelNumber
  
  private String inputFilename = null;
  private Boolean inputIsDeviceOverride = null;
  private int inputOverrideSectorSize = 0;
  private long inputOverrideSize = 0L;
  private long initialBlockSize = -1L;
  private long endBlockSize = -1L;
  private String outputFilename = null;
  private Boolean outputIsDeviceOverride = null;
  private boolean preallocateOutput = false;
  private String mapFilename = null;
  private boolean mapToMemory = false;
  private ArrayList<BlockRange> badBlockEmulatorList = null;

  private long callbackInterval = -1;
  private Runnable callback = null;
  
  private boolean writeToSystemOut = false;

  // Devices
  private BlockDevice inputDevice = null;
  private BlockDevice outputDevice = null;
  private ReadWriteDevice mapFile = null;
  
  // Input device characteristics
  private long inputSize = -1L;
  private long inputSizeSectors = -1L;
  private int inputSectorSizeLog = -1;
  private long inputOffset = 0L;
  private int startLevel = 0;
  private int endLevel = 0;
  
  // Output device characteristics
  private long outputOffset = 0L;
  
  // Map file characteristics
  private long mapFileSizeBytes = -1L;
  private long mapFileDataBytes = -1L;
  private long[] mapFileLevelOffsets = null;
  
  // Number of blocks to divide into in each successive level
  private int childrenPerNode = 4;
  private int childrenPerNodeLog = 2;
  // Number of leaves is the input size rounded up to a power of childrenPerNode to simplify map implementation
  private long numLeaves = -1L;
  private int numLeavesLog = -1;
  private int maxLevelNumber = -1;
  
  // Temp space for updateBlockStatus; only one read from the map is valid at a time because this is global to the instance.
  private byte[] mapFileBytes = null;
  private int[] blockStatusBuffer = null;
  
  // Buffer size
  private int bufferSize = 262144;
  private int bufferSizeSectorsLog = 0;

  // Time to determine when to do callbacks
  private long callbackIntervalStart = 0L;
  
  // State
  private boolean inputIsDevice = false;
  private boolean outputIsDevice = false;
  private int curLevel = 0;
  private long curBlock = 0L;
  
  // Stats
  private long fullReadCount = 0;
  private long partialReadCount = 0;
  private long emptyReadCount = 0;
  private long badReadCount = 0;

  /**
   * Constructs a new ProgressiveCopier
   */
  public ProgressiveCopier()
  {  }
  
  /**
   * Returns whether status messages will be written to System.out. Default is false.
   */
  public boolean getWriteToSystemOut()
  {  return writeToSystemOut;  }
  
  /**
   * Sets whether status messages will be written to System.out.
   */
  public void setWriteToSystemOut(boolean _writeToSystemOut)
  {
    writeToSystemOut = _writeToSystemOut;
    return;
  }
  
  /**
   * Returns the input filename (default null).
   */
  public String getInputFilename()
  {  return inputFilename;  }
  
  /**
   * Sets the input filename. If both an input device and filename are specified, the input device is used.
   */
  public void setInputFilename(String _inputFilename)
  {
    inputFilename = _inputFilename;
    return;
  }
  
  /**
   * Returns the input device (default null).
   */
  public BlockDevice getInputDevice()
  {  return inputDevice;  }
  
  /**
   * Sets the input device. If both an input device and filename are specified, the input device is used.
   */
  public void setInputDevice(BlockDevice _inputDevice)
  {
    inputDevice = _inputDevice;
    return;
  }

  /**
   * Force the input to be treated as a device or not instead of attempting to determine this from an input filename.
   *
   * @param isDevice If true, then treat the input as a device.<br>
   *                 If false, then do not treat the input as a device.
   */
  public void forceInputIsDevice(boolean isDevice)
  {
    inputIsDeviceOverride = Boolean.valueOf(isDevice);
    return;
  }
  
  /**
   * Return whether the input device has been forced and if so, to what status.
   *
   * @return Returns null if not forced, Boolean.TRUE if forced to be a device, and Boolean.FALSE if not forced to be a device.
   */
  public Boolean getForceInputIsDevice()
  {  return inputIsDeviceOverride;  }
  
  /**
   * Lets this object determine if the input is a device (default setting). If the input filename is specified and starts with
   * \\?\ or /dev/ then it is opened as a device.
   */
  public void unforceInputIsDevice()
  {
    inputIsDeviceOverride = null;
    return;
  }
  
  /**
   * Returns whether this object considers the input to be a device. Only valid after the input has been opened.
   */
  public boolean isInputDevice()
  {  return (inputIsDeviceOverride != null) ? inputIsDeviceOverride.booleanValue() : inputIsDevice;  }
  
  /**
   * Returns the overridden sector size of the input. Default for non-devices is 512 bytes.
   */
  public int getInputOverrideSectorSize()
  {  return inputOverrideSectorSize;  }
  
  /**
   * Overrides the input's sector size. Default for non-devices is 512 bytes.
   */
  public void setInputOverrideSectorSize(int _inputOverrideSectorSize)
  {
    inputOverrideSectorSize = _inputOverrideSectorSize;
    return;
  }
  
  /**
   * Returns the overridden size of the input in bytes. A value of 0 means no override.
   */
  public long getInputOverrideSize()
  {  return inputOverrideSize;  }
  
  /**
   * Sets the size of the input in bytes. A value of 0 means no override.
   */
  public void setInputOverrideSize(long _inputOverrideSize)
  {
    inputOverrideSize = _inputOverrideSize;
    return;
  }
  
  /**
   * Returns the block size in bytes that the copy will start from. A value of -1 means to use the default, which means
   * the copier starts with the block size that encompasses the entire input, which is not known until the input is opened.
   */
  public long getInitialBlockSize()
  {  return initialBlockSize;  }
  
  /**
   * Sets the initial block size the copier will start at. A value of -1 means to use the default, which means the copier
   * starts with the block size that encompasses the entire input, which is not known until after the input is opened.
   */
  public void setInitialBlockSize(long _initialBlockSize)
  {
    initialBlockSize = _initialBlockSize;
    return;
  }
  
  /**
   * Returns the smallest block size in bytes that the copier will attempt to read. A value of -1 means to use the default,
   * which means the copier will stop after it finishes blocks at the sector level.
   */
  public long getEndBlockSize()
  {  return endBlockSize;  }
  
  /**
   * Sets the smallest block size in bytes that the copier will attempt to read. A value of -1 means to use the default,
   * which means the copier will stop after it finishes blocks at the sector level.
   */
  public void setEndBlockSize(long _endBlockSize)
  {
    endBlockSize = _endBlockSize;
    return;
  }
  
  /**
   * Returns the output filename (default null).
   */
  public String getOutputFilename()
  {  return outputFilename;  }
  
  /**
   * Sets the output filename. If both an output device and filename are specified, the output device is used.
   */
  public void setOutputFilename(String _outputFilename)
  {
    outputFilename = _outputFilename;
    return;
  }
  
  /**
   * Returns the output device (default null).
   */
  public BlockDevice getOutputDevice()
  {  return outputDevice;  }
  
  /**
   * Sets the output device. If both an output device and filename are specified, the output device is used.
   */
  public void setOutputDevice(BlockDevice _outputDevice)
  {
    outputDevice = _outputDevice;
    return;
  }
  
  /**
   * Force the output to be treated as a device or not instead of attempting to determine this from an output filename.
   *
   * @param isDevice If true, then treat the output as a device.<br>
   *                 If false, then do not treat the output as a device.
   */
  public void forceOutputIsDevice(boolean isDevice)
  {
    outputIsDeviceOverride = Boolean.valueOf(isDevice);
    return;
  }
  
  /**
   * Return whether the output device has been forced and if so, to what status.
   *
   * @return Returns null if not forced, Boolean.TRUE if forced to be a device, and Boolean.FALSE if not forced to be a device.
   */
  public Boolean getForceOutputIsDevice()
  {  return outputIsDeviceOverride;  }

  /**
   * Lets this object determine if the output is a device (default setting). If the output filename is specified and starts with
   * \\?\ or /dev/ then it is opened as a device.
   */
  public void unforceOutputIsDevice()
  {
    outputIsDeviceOverride = null;
    return;
  }

  /**
   * Returns whether this object considers the output to be a device. Only valid after the output has been opened.
   */
  public boolean isOutputDevice()
  {  return (outputIsDeviceOverride != null) ? outputIsDeviceOverride.booleanValue() : outputIsDevice;  }

  /**
   * Returns whether the copier should preallocate the output before starting the copy. Default is false.
   */
  public boolean getPreallocateOutput()
  {  return preallocateOutput;  }
  
  /**
   * Sets whether the copier should preallocate the output before starting the copy. Default is false.
   */
  public void setPreallocateOutput(boolean _preallocateOutput)
  {
    preallocateOutput = _preallocateOutput;
    return;
  }
  
  /**
   * Returns the map filename (default null).
   */
  public String getMapFilename()
  {  return mapFilename;  }
  
  /**
   * Sets the map filename. If both a map file and filename are specified, the map file is used.
   */
  public void setMapFilename(String _mapFilename)
  {
    mapFilename = _mapFilename;
    return;
  }
  
  /**
   * Returns the map file (default null).
   */
  public ReadWriteDevice getMapFile()
  {  return mapFile;  }
  
  /**
   * Sets the map file. If both a map file and filename are specified, the map file is used.
   */
  public void setMapFile(ReadWriteDevice _mapFile)
  {
    mapFile = _mapFile;
    return;
  }

  /**
   * Returns whether the map should be constructed in memory instead of a file. If the map is constructed in memory, it
   * will not persist if the copy finishes or is interrupted.
   */
  public boolean getMapToMemory()
  {  return mapToMemory;  }
  
  /**
   * Sets whether the map should be constructed in memory instead of a file. If the map is constructed in memory, it
   * will not persist if the copy finishes or is interrupted.
   */
  public void setMapToMemory(boolean _mapToMemory)
  {
    mapToMemory = _mapToMemory;
    return;
  }
  
  /**
   * Returns the callback interval in milliseconds. A value of -1 means never callback and 0 means callback after every I/O.
   * Because the callback timer is checked in the same thread as the copier, long I/O operations that block must finish before
   * the callback timer can be checked and can cause the callback to be called late.
   */
  public long getCallbackInterval()
  {  return callbackInterval;  }
  
  /**
   * Sets the callback interval in milliseconds. A value of -1 means never callback and 0 means callback after every I/O.
   */
  public void setCallbackInterval(long _callbackInterval)
  {
    callbackInterval = _callbackInterval;
    return;
  }
  
  /**
   * Returns the Runnable object to be called.
   */
  public Runnable getCallback()
  {  return callback;  }
  
  /**
   * Sets the Runnable object to be called.
   */
  public void setCallback(Runnable _callback)
  {
    callback = _callback;
    return;
  }

  
  /**
   * Returns the number of children per node. This is the number of pieces a block is divided into in succeeding levels. The default value is 4.
   */
  public int getNumChildrenPerNode()
  {
    return childrenPerNode;
  }
  
  /**
   * Sets the number of children per node. This number must be at least 2, and must also be a power of 2.
   * There are several factors this affects. Because the map must cover a virtual device that is at least as large as the actual input device,
   * a large value can lead to an excessively large map.
   * <p>
   * For example, if the sector size of a 1MB device is 512 bytes and the childrenPerNode is 4, the smallest block that covers the entire
   * input device is 512 * 4 * 4 * 4 * 4 * 4 * 4 = 2MB. If the childrenPerNode is set to 32, the smallest block that covers the entire
   * input device is 512 * 32 * 32 * 32 = 16MB, which would result in a much larger map.
   * <p>
   * A smal value increases the number of levels and thus increases the number of times a read attempt may be performed on a bad block,
   * potentially to excessive levels.
   */
  public void setChildrenPerNode(int _childrenPerNode) throws ParamException
  {
    if (_childrenPerNode < 2)
      throw new ParamException("Number of children per node (" + _childrenPerNode + ") must be at least 2");
    int _childrenPerNodeLog = checkPower2(_childrenPerNode);
    if (_childrenPerNodeLog < 0)
      throw new ParamException("Number of children per node (" + _childrenPerNode + ") must be a power of 2");
    childrenPerNode = _childrenPerNode;
    childrenPerNodeLog = _childrenPerNodeLog;
    return;
  }
  
  /**
   * Get the offset into the input at which the data starts.
   */
  public long getInputOffset()
  {  return inputOffset;  }
  
  /**
   * Set the offset into the input at which the data starts.
   */
  public void setInputOffset(long _inputOffset)
  {
    inputOffset = _inputOffset;
    return;
  }
  
  /**
   * Get the offset into the output at which writing should begin. The first sector of the input will be written at this offset (default 0).
   */
  public long getOutputOffset()
  {  return outputOffset;  }
  
  /**
   * Set the offset into the output at which writing should begin.  The first sector of the input will be written at this offset (default 0).
   */
  public void setOutputOffset(long _outputOffset)
  {
    outputOffset = _outputOffset;
    return;
  }
  
  /**
   * Similar to {@link #setChildrenPerNode(int) setChildrenPerNode}, but specifying the base 2 log of childrenPerNode instead.
   */
  public void setChildrenPerNodeLog(int _childrenPerNodeLog)
  {
    childrenPerNodeLog = _childrenPerNodeLog;
    childrenPerNode = 1 << childrenPerNodeLog;
    return;
  }
  
  /**
   * Sets the size of the read buffer (maximum size of read request performed). It must be a power of 2, and larger than the sector size.
   */
  public void setBufferSize(int _bufferSize) throws ParamException
  {
    if (_bufferSize <= 0)
      throw new ParamException("Buffer size must be a positive integer");
    int _bufferSizeLog = checkPower2(_bufferSize);
    if (_bufferSizeLog < 0)
      throw new ParamException("Buffer size must be a power of 2");
    bufferSize = _bufferSize;
    return;
  }
  
  /**
   * Returns the current size of the read buffer
   */
  public int getBufferSize()
  {
    return bufferSize;
  }
  
  /**
   * Returns the current level that is being copied.
   */
  public int getCurLevel()
  {
    return curLevel;
  }
  
  /**
   * Returns the block number in the current level that is being copied.
   */
  public long getCurBlock()
  {
    return curBlock;
  }
  
  /**
   * Returns the number of full reads performed (statistic).
   */
  public long getFullReadCount()
  {
    return fullReadCount;
  }
  
  /**
   * Returns the number of partial reads performed (statistic).
   */
  public long getPartialReadCount()
  {
    return partialReadCount;
  }
  
  /**
   * Returns the number of empty reads performed (statistic).
   */
  public long getEmptyReadCount()
  {
    return emptyReadCount;
  }
  
  /**
   * Returns the number of bad reads performed (statistic).
   */
  public long getBadReadCount()
  {
    return badReadCount;
  }
  
  /**
   * Returns the size of the input (may not be known until the copy starts).
   */
  public long getInputSize()
  {
    return inputSize;
  }
  
  /**
   * Returns the size of the input in sectors (may not be known until the copy starts).
   */
  public long getInputSizeSectors()
  {
    return inputSizeSectors;
  }
  
  /**
   * Returns the base 2 logarithm of the sector size of the input (may not be known until the copy starts).
   */
  public int getInputSectorSizeLog()
  {
    return inputSectorSizeLog;
  }
  
  /**
   * Returns the level at which the copy will start
   */
  public int getStartLevel()
  {
    return startLevel;
  }
  
  /**
   * Returns the level at which the copy will end
   */
  public int getEndLevel()
  {
    return endLevel;
  }
  
  /**
   * Returns the number of virtual sectors in the input device. This is a power of 2,
   * and will be the same or larger than the number of actual sectors in the input device.
   */
  public long getNumLeaves()
  {
    return numLeaves;
  }
  
  /**
   * Returns the base 2 logarithm of the number of leaves in the map. The number of leaves corresponds to the number
   * of virtual sectors in the map.
   */
  public int getNumLeavesLog()
  {
    return numLeavesLog;
  }
  
  /**
   * Returns the largest level number (the level at which the block size is equal to 1 sector).
   */
  public int getMaxLevelNumber()
  {
    return maxLevelNumber;
  }
  
  /**
   * Reads and appends a list of bad sector runs from the given Reader. The input device will be wrapped in a
   * {@link org.blinkenbyte.io.BadBlockEmulator BadBlockEmulator} to prevent the sectors from being read from.
   * <p>
   * The format of sector lists is specified in the class description.
   */
  public void emulateBadSectors(Reader r) throws Exception
  {
    if (badBlockEmulatorList == null)
      badBlockEmulatorList = new ArrayList<BlockRange>();
    BlockRange.readList(r, badBlockEmulatorList);
    BlockRange.coalesce(badBlockEmulatorList);
    return;
  }
  
  /**
   * Appends the given list of blocks to the bad sector list. The input device will be wrapped in a
   * {@link org.blinkenbyte.io.BadBlockEmulator BadBlockEmulator} to prevent the sectors from being read from.
   */
  public void emulateBadSectors(List<BlockRange> listToAppend)
  {
    if (badBlockEmulatorList == null)
      badBlockEmulatorList = new ArrayList<BlockRange>();
    badBlockEmulatorList.addAll(listToAppend);
    BlockRange.coalesce(badBlockEmulatorList);
    return;
  }
  
  /**
   * Appends the given block to the bad sector list. The input device will be wrapped in a
   * {@link org.blinkenbyte.io.BadBlockEmulator BadBlockEmulator} to prevent the sectors from being read from.
   */
  public void addEmulatedBadSector(BlockRange position)
  {
    if (badBlockEmulatorList == null)
      badBlockEmulatorList = new ArrayList<BlockRange>();
    badBlockEmulatorList.add(position);
    BlockRange.coalesce(badBlockEmulatorList);
    return;
  }
  
  /**
   * Clears the list of emulated bad sectors.
   */
  public void clearEmulateBadSectorList()
  {
    badBlockEmulatorList.clear();
    return;
  }
  
  /**
   * Opens the input, output, and map files/devices as needed and gathers basic information from them. This must be used
   * if performCopy needs to be called. If used, openMapOnly() must not be called.
   */
  public void open() throws Exception
  {
    openInput();
    openOutput();
    openMap(false);
    return;
  }
  
  /**
   * Opens the map and prepares it to be read or cleared. This must only be used when only the map needs to be manipulated,
   * as with writeIncompleteSectorsList and clearMap. If used, open() must not be called.
   */
  public void openMapOnly() throws Exception
  {
    openMap(true);
    return;
  }
  
  /**
   * Performs the progressive copy.
   */
  public void performCopy() throws IOException
  {
    if (writeToSystemOut)
      System.out.println("Starting copy");
    byte[] buffer = new byte[1 << (inputSectorSizeLog + bufferSizeSectorsLog)];
    int bufferSizeSectors = 1 << bufferSizeSectorsLog;
    callbackIntervalStart = System.currentTimeMillis();
    for (curLevel = startLevel; curLevel <= endLevel; curLevel++)
    {
      long blocksInCurLevel = calcNumBlocksForLevel(curLevel);
      long curLevelBlockSizeSectors = calcBlockSizeInSectors(curLevel);
      long curLevelBlockSizeBytes = curLevelBlockSizeSectors << inputSectorSizeLog;
      inputDevice.seek(inputOffset);
      outputDevice.seek(outputOffset);
      // Initial value to determine how many levels up to check for completed nodes
      // The -1 is done because the root node does not need to be checked.
      long prevBlockNumber = blocksInCurLevel - 1;
      boolean lastSkipped = false;
      for (curBlock = 0; curBlock < blocksInCurLevel; curBlock++)
      {
        lastSkipped = true;
        boolean blockUnread = (getBlockStatus(curLevel, curBlock) == STATUS_UNREAD);
        // See how far up we need to check for completed blocks based on the last check that was done
        // The bits representing the children of the parent node are shifted out
        long higherBlocksToCheck = (curBlock ^ prevBlockNumber) >>> childrenPerNodeLog;
        prevBlockNumber = curBlock;
        int levelCheckOffset = 1;
        int levelCheckOffsetMult = childrenPerNodeLog;
        int highestCompletedLevelOffsetMult = -1;
        while (higherBlocksToCheck != 0)
        {
          if (getBlockStatus(curLevel - levelCheckOffset, curBlock >>> levelCheckOffsetMult) == STATUS_COMPLETE)
            highestCompletedLevelOffsetMult = levelCheckOffsetMult;
          levelCheckOffset++;
          levelCheckOffsetMult += childrenPerNodeLog;
          higherBlocksToCheck >>>= childrenPerNodeLog;
        }
        // Check if a higher level block is complete
        if (highestCompletedLevelOffsetMult > 0)
        {
          curBlock |= (1 << highestCompletedLevelOffsetMult) - 1;
          // Equivalent to: curBlock = (((curBlock >>> highestCompletedLevelOffsetMult) + 1) << highestCompletedLevelOffsetMult) - 1;
          // The -1 is because the for() loop will increment curBlock
          checkCallback();
          continue;
        }
        // If the block has been touched, skip it
        if (!blockUnread)
        {
          checkCallback();
          continue;
        }

        long curBlockSectorStart = calcSectorNumber(curLevel, curBlock);
        long sectorOffset = 0L;

        if (lastSkipped)
        {
          if (curBlockSectorStart < inputSizeSectors)
          {
            inputDevice.seek(inputOffset + (curBlockSectorStart << inputSectorSizeLog));
            outputDevice.seek(outputOffset + (curBlockSectorStart << inputSectorSizeLog));
          }
        }
        lastSkipped = false;
        while (sectorOffset < curLevelBlockSizeSectors)
        {
          checkCallback();
          long virtualReadSizeSectors = curLevelBlockSizeSectors - sectorOffset;
          if ((curBlockSectorStart + sectorOffset) >= inputSizeSectors)
          {
            updateMap(curLevel, curBlock, sectorOffset, virtualReadSizeSectors);
            sectorOffset += virtualReadSizeSectors;
            continue;
          }
          if (virtualReadSizeSectors > bufferSizeSectors)
            virtualReadSizeSectors = bufferSizeSectors;
          int physicalReadSizeSectors = (int)virtualReadSizeSectors;
          if ((curBlockSectorStart + sectorOffset + virtualReadSizeSectors) > inputSizeSectors)
            physicalReadSizeSectors = (int)(inputSizeSectors - (curBlockSectorStart + sectorOffset));
          int numRequestedBytes = physicalReadSizeSectors << inputSectorSizeLog;
          int numBytesRead = -1;
          try
          {
            numBytesRead = inputDevice.read(buffer, 0, numRequestedBytes);
          }
          catch (Exception e)
          {
            badReadCount++;
// Optional logging here
          }
          int numSectorsRead = (numBytesRead < 0) ? 0 : (numBytesRead >>> inputSectorSizeLog);
          // If fully read
          if (numSectorsRead == physicalReadSizeSectors)
          {
            outputDevice.write(buffer, 0, numBytesRead);
            updateMap(curLevel, curBlock, sectorOffset, virtualReadSizeSectors);
            sectorOffset += virtualReadSizeSectors;
            fullReadCount++;
            continue;
          }
          // If some data read
          if (numSectorsRead > 0)
          {
            outputDevice.write(buffer, 0, numBytesRead);
            updateMap(curLevel, curBlock, sectorOffset, numSectorsRead);
            sectorOffset += numSectorsRead;
            partialReadCount++;
            continue;
          }
          if (numBytesRead == 0)
            emptyReadCount++;
          // If at a leaf, mark the leaf as bad
          if (curLevel == maxLevelNumber)
          {
            updateBlockStatusCascade(curLevel, curBlock, STATUS_ERROR);
          }
          else
          {
            // If at the start of the block, mark the block at the level closest to the size of the read as partial
            // This is done so that all blocks aligned on this boundary will not be read
            // (if the first sector is bad, this will prevent an attempted read of the sector at every single level)
            if (sectorOffset == 0)
            {
              int targetLevelOffset = 0;
              long targetLevelSectorSize = 1;
              while (targetLevelSectorSize < physicalReadSizeSectors)
              {
                targetLevelOffset++;
                targetLevelSectorSize <<= childrenPerNodeLog;
              }
              if ((targetLevelOffset + curLevel) > maxLevelNumber)
                throw new RuntimeException("Internal error: Special case for marking blocks bad at the start of a read failed sanity check!");
              int oldBlockStatus = getBlockStatus(maxLevelNumber - targetLevelOffset, curBlock << ((maxLevelNumber - targetLevelOffset - curLevel) * childrenPerNodeLog));
              if (oldBlockStatus == STATUS_UNREAD)
                updateBlockStatusCascade(maxLevelNumber - targetLevelOffset, curBlock << ((maxLevelNumber - targetLevelOffset - curLevel) * childrenPerNodeLog), STATUS_PARTIAL);
            }
          }
          // Seek to next block
          long nextPos = curBlockSectorStart + curLevelBlockSizeSectors;
          if (nextPos < inputSizeSectors)
          {
            inputDevice.seek(inputOffset + (nextPos << inputSectorSizeLog));
            outputDevice.seek(outputOffset + (nextPos << inputSectorSizeLog));
          }
          break;
        }
      }
    }
    return;
  }
  
  /**
   * Closes the input, output, and map files/devices if they were opened by this class. If the device/file was passed
   * directly instead of being opened by this class, they will not be closed.
   */
  public void close()
  {
    if ((mapFilename == null) && (mapFile != null))
    {
      try
      {  mapFile.close();  }
      catch (Exception e)
      {  }
      mapFile = null;
    }
    if ((outputFilename == null) && (outputDevice != null))
    {
      try
      {  outputDevice.close();  }
      catch (Exception e)
      {  }
      outputDevice = null;
    }
    if ((inputFilename == null) && (inputDevice != null))
    {
      try
      {  inputDevice.close();  }
      catch (Exception e)
      {  }
      inputDevice = null;
    }
    return;
  }
  
  // Checks if enough time has passed that the callback needs to be called.
  private void checkCallback()
  {
    if (callback == null)
      return;
    if (callbackInterval < 0L)
      return;
    // A callback interval of 0 means to run the callback every time
    if ((System.currentTimeMillis() - callbackIntervalStart) >= callbackInterval)
    {
      callback.run();
      callbackIntervalStart = System.currentTimeMillis();
    }
    return;
  }
  
  // Returns the status of the specified block.
  private int getBlockStatus(int blockLevel, long blockNumber) throws IOException
  {
    long sectorNumber = calcSectorNumber(blockLevel, blockNumber);
    long mapBlockNumber = mapFileLevelOffsets[blockLevel] + blockNumber;
    long mapBlockOffset = mapBlockNumber >> 2;
    mapFileSeek(mapBlockOffset);
    int mapByte = mapFile.read();
    if (mapByte < 0)
      throw new IOException("Unable to read byte from map file at offset " + (MAPFILE_HEADER_SIZE + mapBlockOffset) + " (blockLevel=" + blockLevel + ", blockNumber=" + blockNumber + ")");
    return (mapByte >>> ((mapBlockNumber & 3) << 1)) & 3;
  }
  
  // Updates the block status specified and returns the status the parent should have (except level 0)
  // If the status is less than zero, then the parent status is updated but the requested block status is not updated (perform query).
  // If an array of statuses is passed, ignore newStatus and update/query the requested siblings
  private int updateBlockStatus(int blockLevel, long blockNumber, int newStatus, int[] newStatuses) throws IOException
  {
    long baseBlockNumber = blockNumber & ~(childrenPerNode - 1);
    long sectorsPerBlock = calcBlockSizeInSectors(blockLevel);
    long sectorNumber = calcSectorNumber(blockLevel, baseBlockNumber);
    long mapBlockNumber = mapFileLevelOffsets[blockLevel] + baseBlockNumber;
    long mapBlockOffset = mapBlockNumber >> 2;
    int mapByte = 0;
    int mapFileByteIndex = 0;
    readMapFileBytes(mapBlockOffset, mapFileBytes, 0, mapFileBytes.length);
    mapByte = mapFileBytes[0];
    int byteShift = (int)(mapBlockNumber & 3) << 1;

    int howManyEmpty = 0;
    int howManyPartial = 0;
    int howManyComplete = 0;
    boolean isLeaf = (blockLevel == maxLevelNumber);
    boolean isRoot = (blockLevel == 0);
    for (int i = 0; i < (isRoot ? 1 : childrenPerNode); i++)
    {
      // Check if another byte needs to be read
      if (byteShift == 8)
      {
        mapBlockOffset++;
        mapByte = mapFileBytes[++mapFileByteIndex];
        byteShift = 0;
      }
      if (newStatuses != null)
        newStatus = newStatuses[i];
      if (((newStatuses != null) || ((baseBlockNumber | i) == blockNumber)) && (newStatus >= 0))
      {
        mapByte = (mapByte & ~(3 << byteShift)) | (newStatus << byteShift);
        mapFileSeek(mapBlockOffset);
        mapFile.write(mapByte);
        blockStatusBuffer[i] = newStatus;
      }
      else
        blockStatusBuffer[i] = (mapByte >>> byteShift) & 3;
      sectorNumber += sectorsPerBlock;
      byteShift += 2;

      if (blockStatusBuffer[i] == STATUS_UNREAD)
        howManyEmpty++;
      else
      {
        if (blockStatusBuffer[i] == STATUS_COMPLETE)
          howManyComplete++;
        else
          howManyPartial++;
      }
    }
    if (isRoot)
      return blockStatusBuffer[0];
    
    if (howManyPartial > 0)
      return STATUS_PARTIAL;
    if (howManyEmpty == 0)
      return STATUS_COMPLETE;
    if (howManyComplete == 0)
      return STATUS_UNREAD;
    return STATUS_PARTIAL;
  }
  
  // Updates the block status of the specified block and cascades to parent blocks as needed.
  private void updateBlockStatusCascade(int blockLevel, long blockNumber, int newStatus) throws IOException
  {
    while (true)
    {
      newStatus = updateBlockStatus(blockLevel, blockNumber, newStatus, null);
      if (blockLevel == 0)
        break;
      blockLevel--;
      blockNumber >>>= childrenPerNodeLog;
      int curParentStatus = getBlockStatus(blockLevel, blockNumber);
      if (newStatus == curParentStatus)
        break;
    }
    return;
  }
  
  // Updates the map for the specified run of sectors in the given block.
  private void updateMap(int blockLevel, long blockNumber, long blockSectorOffset, long blockSectorCount) throws IOException
  {
    // Assume that blockSectorOffset/blockSectorCount will not exceed the bounds of the current block
    // Also assumes that if the end of the block is being read, the start of the block has already been read
    if (blockLevel == maxLevelNumber)
    {
      updateBlockStatusCascade(blockLevel, blockNumber, STATUS_COMPLETE);
      return;
    }
    long curBlockSize = calcBlockSizeInSectors(blockLevel);
    long childBlockSize = curBlockSize >>> childrenPerNodeLog;
    long blockSectorEnd = blockSectorOffset + blockSectorCount;
    if (blockSectorEnd > curBlockSize)
      throw new RuntimeException("Internal error: updateMap() called with invalid parameters: blockLevel=" + blockLevel + " blockNumber=" + blockNumber + " blockSectorOffset=" + blockSectorOffset + " blockSectorCount=" + blockSectorCount);
    if (blockSectorEnd == curBlockSize)
    {
      updateBlockStatusCascade(blockLevel, blockNumber, STATUS_COMPLETE);
      return;
    }
    long curChildStartOffset = 0L;
    long curChildEndOffset = 0L;
    for (int i = 0; i < childrenPerNode; i++)
    {
      curChildStartOffset = curChildEndOffset;
      curChildEndOffset += childBlockSize;
      // Does the child overlap with the range passed to this method?
      if ((blockSectorEnd > curChildStartOffset) && (blockSectorOffset < curChildEndOffset))
      {
        // AND the two blocks
        // Pick the larger start offset
        long updateStartOffset = (curChildStartOffset > blockSectorOffset) ? curChildStartOffset : blockSectorOffset;
        // Pick the smaller end offset
        long updateEndOffset = (curChildEndOffset < blockSectorEnd) ? curChildEndOffset : blockSectorEnd;
        updateMap(blockLevel + 1, (blockNumber << childrenPerNodeLog) | i, updateStartOffset - curChildStartOffset, updateEndOffset - updateStartOffset);
      }
    }
    return;
  }
  
  // Opens the map and computes and verifies various internal values.
  private void openMap(boolean mapOnly) throws Exception
  {
    boolean isNewMap = false;
    if (mapFile == null)
    {
      if (mapToMemory)
      {
        mapFile = new ReadWriteByteArray();
        isNewMap = true;
      }
      else
      {
        mapFile = new ReadWriteFile(mapFilename, "rwd");
        isNewMap = (mapFile.length() == 0);
      }
    }
    else
      isNewMap = (mapFile.length() == 0);
    
    if (mapOnly && isNewMap)
      throw new Exception("Map is a new or empty file");
    
    // Check header
    if (!isNewMap)
    {
      byte[] mapFileHeader = new byte[MAPFILE_HEADER_SIZE];
      try
      {
        mapFile.seek(0L);
        IOUtil.readFully(mapFile, mapFileHeader);
      }
      catch (Exception e)
      {
        throw new Exception("Unable to read header from map file");
      }
      int ind = 0;
      for (ind = 0; ind < MAPFILE_HEADER_MARKER.length; ind++)
        if (mapFileHeader[ind] != MAPFILE_HEADER_MARKER[ind])
          throw new Exception("Map file's header is invalid");
      int versionNumber = readLittleEndian(mapFileHeader, ind, 4);
      ind += 4;
      if (versionNumber != CURRENT_MAP_VERSION)
        throw new Exception("Invalid version number in map file (found " + versionNumber + ", expected " + CURRENT_MAP_VERSION + ")");
      int _childrenPerNode = readLittleEndian(mapFileHeader, ind, 4);
      ind += 4;
      if (mapOnly)
        setChildrenPerNode(_childrenPerNode);
      else
        if (_childrenPerNode != childrenPerNode)
          throw new Exception("Map file has a different number of children per node (found " + _childrenPerNode + ", expected " + childrenPerNode + ")");
      int _maxLevelNumber = readLittleEndian(mapFileHeader, ind, 4);
      if (mapOnly)
      {
        maxLevelNumber = _maxLevelNumber;
        numLeavesLog = maxLevelNumber * childrenPerNodeLog;
        numLeaves = 1 << numLeavesLog;
        inputSizeSectors = numLeaves;
      }
      else
        if (maxLevelNumber != _maxLevelNumber)
          throw new Exception("Map file appears to be for a different input size (map covers " + calcNumBlocksForLevel(_maxLevelNumber) + " blocks, current input's map needs to cover " + calcNumBlocksForLevel(maxLevelNumber) + ")");
    }
    else
    {
      mapFile.write(MAPFILE_HEADER_MARKER);
      writeLittleEndian(mapFile, CURRENT_MAP_VERSION);
      writeLittleEndian(mapFile, childrenPerNode);
      writeLittleEndian(mapFile, maxLevelNumber);
    }

    mapFileLevelOffsets = new long[maxLevelNumber + 1];
    long numMapBlocks = 1;
    mapFileLevelOffsets[0] = 0;
    for (int i = 1; i <= maxLevelNumber; i++)
    {
      mapFileLevelOffsets[i] = numMapBlocks;
      numMapBlocks = (numMapBlocks << childrenPerNodeLog) | 1;
    }
    mapFileDataBytes = ((numMapBlocks + 3) >> 2);
    mapFileSizeBytes = MAPFILE_HEADER_SIZE + mapFileDataBytes;		// map file header, 4 block statuses per byte
    if (isNewMap && (mapFile instanceof ReadWriteByteArray))
      ((ReadWriteByteArray)mapFile).setLength((int)mapFileSizeBytes);

    mapFileBytes = new byte[((childrenPerNode < 4) ? 1 : (childrenPerNode >>> 2)) + 1];
    blockStatusBuffer = new int[childrenPerNode];

    if (mapFile.length() >= mapFileSizeBytes)
      return;
    if (writeToSystemOut)
      System.out.println("Allocating space for map file...");
    byte[] buffer = new byte[1048576];
    while (mapFileSizeBytes > 0)
    {
      int bytesToWrite = (mapFileSizeBytes > buffer.length) ? buffer.length : (int)mapFileSizeBytes;
      mapFile.write(buffer, 0, bytesToWrite);
      mapFileSizeBytes -= bytesToWrite;
    }
    return;
  }
  
  private int readLittleEndian(byte[] b, int offset, int len)
  {
    int n = 0;
    for (int i = len - 1; i >= 0; i--)
      n = (n << 8) | ((int)b[offset + i] & 0xFF);
    return n;
  }
  
  private void writeLittleEndian(int n, byte[] b, int offset, int len)
  {
    for (int i = 0; i < len; i++)
    {
      b[offset + i] = (byte)n;
      n >>>= 8;
    }
    return;
  }
  
  private void writeLittleEndian(ReadWriteDevice rwd, int n) throws IOException
  {
    byte[] b = new byte[4];
    writeLittleEndian(n, b, 0, 4);
    rwd.write(b);
    return;
  }
  
  private void mapFileSeek(long n) throws IOException
  {
    mapFile.seek(MAPFILE_HEADER_SIZE + n);
    return;
  }
  
  /**
   * Writes a list of incomplete sectors to the console or specified file. The format is specified in the class description.
   *
   * @param outputIncompleteSectorsFilename Name of the file to write the list of incomplete sectors. If the name is null,
   *                                        the list is written to System.out.
   */
  public void writeIncompleteSectorsList(String outputIncompleteSectorsFilename) throws IOException
  {
    BufferedWriter bw = null;
    try
    {
      if (outputIncompleteSectorsFilename == null)
        bw = new BufferedWriter(new OutputStreamWriter(System.out));
      else
        bw = new BufferedWriter(new FileWriter(outputIncompleteSectorsFilename));
      IncompleteBlockIterator incompleteBlockIterator = new IncompleteBlockIterator();
      while (incompleteBlockIterator.hasNext())
      {
        BlockRange position = incompleteBlockIterator.next();
        bw.write(position.getLine());
        bw.newLine();
      }
      return;
    }
    catch (IncompleteBlockIteratorException ibie)
    {
      throw (IOException)ibie.getCause();
    }
    finally
    {
      if (bw != null)
      {
        if (outputIncompleteSectorsFilename == null)
          bw.flush();
        else
        {
          try
          {  bw.close();  }
          catch (Exception e)
          {  }
        }
        bw = null;
      }
    }
  }
  
  /**
   * Returns a list of incomplete sectors.
   */
  public ArrayList<BlockRange> getIncompleteSectorsList() throws IOException
  {
    ArrayList<BlockRange> list = new ArrayList<BlockRange>();
    IncompleteBlockIterator incompleteBlockIterator = new IncompleteBlockIterator();
    while (incompleteBlockIterator.hasNext())
      list.add(incompleteBlockIterator.next());
    return list;
  }
  
  // Class used to iterate the list of incomplete blocks using the map.
  private class IncompleteBlockIterator implements Iterator<BlockRange>
  {
  
    private BlockRange nextPosition = null;
    private long numBlocks = numLeaves;
    private long prevBlockNumber = numBlocks - 1;
    private long firstUnreadBlock = -1L;
    private long curBlock = 0;
    
    public IncompleteBlockIterator() throws IOException
    {
      setNextPosition();
    }
    
    public boolean hasNext()
    {  return (nextPosition != null);  }
    
    public BlockRange next()
    {
      if (nextPosition == null)
        throw new NoSuchElementException();
      BlockRange retPos = nextPosition;
      try
      {
        setNextPosition();
      }
      catch (IOException ioe)
      {
        throw new IncompleteBlockIteratorException(ioe);
      }
      return retPos;
    }
    
    public void remove()
    {
      throw new UnsupportedOperationException();
    }
    
    private void setNextPosition() throws IOException
    {
      nextPosition = null;
      while (curBlock < numBlocks)
      {
        int status = getBlockStatus(maxLevelNumber, curBlock);
        boolean blockUnread = (status == STATUS_UNREAD);
        boolean blockComplete = (status == STATUS_COMPLETE);
        // See how far up we need to check for completed blocks based on the last check that was done
        // The bits representing the children of the parent node are shifted out
        long higherBlocksToCheck = (curBlock ^ prevBlockNumber) >>> childrenPerNodeLog;
        prevBlockNumber = curBlock;
        int levelCheckOffset = 1;
        int levelCheckOffsetMult = childrenPerNodeLog;
        int highestCompletedLevelOffsetMult = -1;
        int highestUnreadLevelOffsetMult = -1;
        while (higherBlocksToCheck != 0)
        {
          status = getBlockStatus(maxLevelNumber - levelCheckOffset, curBlock >>> levelCheckOffsetMult);
          if (status == STATUS_COMPLETE)
            highestCompletedLevelOffsetMult = levelCheckOffsetMult;
          else
            if (status == STATUS_UNREAD)
              highestUnreadLevelOffsetMult = levelCheckOffsetMult;
          levelCheckOffset++;
          levelCheckOffsetMult += childrenPerNodeLog;
          higherBlocksToCheck >>>= childrenPerNodeLog;
        }
        // Check if a higher level block is complete
        if (blockComplete || (highestCompletedLevelOffsetMult > 0))
        {
          long origCurBlock = curBlock;
          if (highestCompletedLevelOffsetMult > 0)
            curBlock |= (1 << highestCompletedLevelOffsetMult) - 1;
          curBlock++;
          if (firstUnreadBlock >= 0)
          {
            nextPosition = new BlockRange(firstUnreadBlock, origCurBlock - firstUnreadBlock);
            firstUnreadBlock = -1;
            return;
          }
          continue;
        }
        if (firstUnreadBlock < 0)
          firstUnreadBlock = curBlock;
        if (blockUnread && (highestUnreadLevelOffsetMult > 0))
          curBlock |= (1 << highestUnreadLevelOffsetMult) - 1;
        curBlock++;
      }
      if (firstUnreadBlock < 0)
        return;
      nextPosition = new BlockRange(firstUnreadBlock, numBlocks - firstUnreadBlock);
      firstUnreadBlock = -1;
      return;
    }
    
  }
  
  // Class used to wrap IOException in a RuntimeException so it can be smuggled out of an Iterator.
  private static class IncompleteBlockIteratorException extends RuntimeException
  {
    public IncompleteBlockIteratorException(IOException ioe)
    {  super(ioe);  }
  }

  /**
   * Clears entries from the map file. The clearing operation starts at the specified level and range of blocks, and continues 
   * to smaller and smaller blocks, eventually reaching the sector level. The range of blocks is limited by the range of sectors
   * encompassed by the original range of blocks.
   *
   * @param mapClearStartingLevel The level at which to start the clear operation. Level 0 is the top level, containing the single
   *                              block that encompasses the entire drive.
   * @param mapClearBlockStart    The block at which to start clearing in the specified level.
   * @param mapClearBlockCount    The number of blocks to clear in the specified level. If less than zero, all remaining blocks in
                                  the level beginning at mapClearBlockStart are cleared.
   */
  public void clearMapBlocks(int mapClearStartingLevel, long mapClearBlockStart, long mapClearBlockCount) throws IOException
  {
    if (mapClearStartingLevel < 0)
      throw new RuntimeException("Starting level below 0 is invalid");
    if (mapClearStartingLevel > maxLevelNumber)
      throw new RuntimeException("Starting level above " + maxLevelNumber + " is invalid");
    long blocksInCurLevel = calcNumBlocksForLevel(mapClearStartingLevel);
    if (mapClearBlockStart >= blocksInCurLevel)
      throw new RuntimeException("Starting block number is greater or equal to the number of blocks in this level (" + blocksInCurLevel + ")");
    if (mapClearBlockCount == 0)
      return;
    if (mapClearBlockCount < 0)
      mapClearBlockCount = blocksInCurLevel - mapClearBlockStart;
    else
      if ((mapClearBlockStart + mapClearBlockCount) > blocksInCurLevel)
        throw new RuntimeException("Blocks to be cleared extends past the end of the current level; maximum given current parameters is " + (blocksInCurLevel - mapClearBlockStart));
    if (writeToSystemOut)
      System.out.println("Starting map clear from level " + mapClearStartingLevel + " block #" + mapClearBlockStart + ", for " + mapClearBlockCount + " blocks");
    long curBlocksToClearStart = mapClearBlockStart;
    long curBlocksToClearCount = mapClearBlockCount;
    for (int curClearLevel = mapClearStartingLevel; curClearLevel <= maxLevelNumber; curClearLevel++)
    {
      clearMapBlockRange(curClearLevel, curBlocksToClearStart, curBlocksToClearCount);
      curBlocksToClearStart <<= childrenPerNodeLog;
      curBlocksToClearCount <<= childrenPerNodeLog;
    }
    return;
  }
  
  /**
   * Clears entries from the map file. The clearing operation marks the specified range of sectors as unread, propagating the
   * change through all levels.
   *
   * @param clearSectorStart The sector at which to start clearing map entries.
   * @param clearSectorCount The number of sector map entries to clear. If the count is negative, it will be computed as all
   *                         remaining sectors in the map.
   */
  public void clearMapSectors(long clearSectorStart, long clearSectorCount) throws IOException
  {
    if (clearSectorStart < 0)
      throw new RuntimeException("Starting sector can not be negative");
    if (clearSectorStart >= numLeaves)
      throw new RuntimeException("Starting sector is larger than the number of sectors in the map");
    if (clearSectorCount == 0)
      return;
    if (clearSectorCount < 0)
      clearSectorCount = numLeaves - clearSectorStart;
    if ((clearSectorStart + clearSectorCount) > numLeaves)
      throw new RuntimeException("Sectors to be cleared extends past the end of the map; maximum given current parameters is " + (numLeaves - clearSectorStart));

    long clearSectorEnd = clearSectorStart + clearSectorCount;
    // Loop through all the levels.
    for (int clearLevel = 0; clearLevel <= maxLevelNumber; clearLevel++)
    {
      long numSectorsPerBlock = calcBlockSizeInSectors(clearLevel);
      long mask = numSectorsPerBlock - 1;
      // Determine the range of affected blocks
      long startingBlock = calcBlockNumber(clearLevel, clearSectorStart);
      long endingBlock = calcBlockNumber(clearLevel, clearSectorEnd);
      boolean startBlockIsPartial = ((clearSectorStart & mask) != 0);
      boolean endBlockIsPartial = ((clearSectorEnd & mask) != 0);
      long fullClearStartingBlock = startingBlock + (startBlockIsPartial ? 1 : 0);
      long fullClearEndingBlock = endingBlock;
      if (fullClearStartingBlock < fullClearEndingBlock)
        clearMapBlockRange(clearLevel, fullClearStartingBlock, fullClearEndingBlock - fullClearStartingBlock);
      // Now to check the partials
      if (startBlockIsPartial || endBlockIsPartial)
      {
        long[] partialList = null;
        if (startingBlock < endingBlock)
          partialList = new long[] { startingBlock, endingBlock };
        else
          partialList = new long[] { startingBlock };
        for (int i = 0; i < partialList.length; i++)
        {
          long curBlockNumber = partialList[i];

          // Read the block status
          int curStatus = getBlockStatus(clearLevel, curBlockNumber);
          if (curStatus != STATUS_UNREAD)
          {
            int[] newStatuses = null;
            if (curStatus == STATUS_COMPLETE)
            {
              // The STATUS_COMPLETE status needs to be pushed down to the children since the parent will no longer be STATUS_COMPLETE
              // All children should be marked as STATUS_COMPLETE unless they are contained within the sector range.
              // Any that are partial overlaps will be noticed in the next level. They should not be set to partial at this stage,
              // so that their STATUS_COMPLETE can be pushed down to their children.
              newStatuses = new int[childrenPerNode];
              long childSectorNumberStart = calcSectorNumber(clearLevel, curBlockNumber);
              for (int j = 0; j < childrenPerNode; j++)
              {
                long childSectorNumberEnd = childSectorNumberStart + (numSectorsPerBlock >>> childrenPerNodeLog);
                if ((childSectorNumberStart >= clearSectorStart) && (childSectorNumberEnd <= clearSectorEnd))
                  newStatuses[j] = STATUS_UNREAD;
                else
                  newStatuses[j] = STATUS_COMPLETE;
                childSectorNumberStart = childSectorNumberEnd;
              }
            }
            
            // The status should be updated to partial or empty, depending on the children.
            // Since updateBlockStatus does something like this, we'll use it to query the children
            // Pick the first child and query what the parent status should be. We are guaranteed to not
            // be on the last level at this point because the sector level can't have partial blocks.
            // If some children need to be updated, that will be done here via newStatuses.
            int newStatus = updateBlockStatus(clearLevel + 1, curBlockNumber << childrenPerNodeLog, -1, newStatuses);
            // If the current block was STATUS_COMPLETE, force the new status to be STATUS_PARTIAL.
            if (newStatuses != null)
              newStatus = STATUS_PARTIAL;
            // If the status is different, we actually have to update
            if (curStatus != newStatus)
            {
              // Now actually update the block. We don't need to cascade updates since we'll be checking the next level.
              updateBlockStatus(clearLevel, curBlockNumber, newStatus, null);
            }
          }
        }
      }
    }
    return;
  }

  /**
   * Clears entries from the map file. The clearing operation marks the specified range of sectors as unread, and propagates the
   * change to higher levels (larger blocks), eventually reaching the level 0 (the single block that encompasses the entire input).
   *
   * @param clearSectorStart The sector at which to start clearing map entries.
   * @param clearSectorCount The number of sector map entries to clear. If it extends past the end of the map, it will be limited
   *                         to the end of the map.
   */
/*  
  public void clearMapSectors(long clearSectorStart, long clearSectorCount) throws IOException
  {
    if (clearSectorStart < 0)
      throw new RuntimeException("Starting sector can not be negative");
    if (clearSectorStart >= numLeaves)
      throw new RuntimeException("Starting sector is larger than the number of sectors in the map");
    if (clearSectorCount <= 0)
      throw new RuntimeException("Sector count must be greater than 0");
    if ((clearSectorStart + clearSectorCount) > numLeaves)
      clearSectorCount = numLeaves - clearSectorStart;
    
    long clearSectorEnd = clearSectorStart + clearSectorCount;
    // Loop through all the levels.
    for (int clearLevel = maxLevelNumber; clearLevel >= 0; clearLevel--)
    {
      long numSectorsPerBlock = calcBlockSizeInSectors(clearLevel);
      long mask = numSectorsPerBlock - 1;
      // Determine the range of affected blocks
      long startingBlock = clearSectorStart >>> (childrenPerNodeLog * clearLevel);
      long endingBlock = (clearSectorEnd + mask) >>> (childrenPerNodeLog * clearLevel);
      boolean startBlockIsPartial = ((clearSectorStart & mask) != 0);
      boolean endBlockIsPartial = ((clearSectorEnd & mask) != 0);

      long fullClearStartingBlock = startingBlock + (startBlockIsPartial ? 1 : 0);
      long fullClearEndingBlock = endingBlock - (endBlockIsPartial ? 1 : 0);
      if (fullClearStartingBlock < fullClearEndingBlock)
        clearMapBlockRange(clearLevel, fullClearStartingBlock, fullClearEndingBlock - fullClearStartingBlock);
      // Now to check the partials
      if (startBlockIsPartial || endBlockIsPartial)
      {
        endingBlock--;
        long[] partialList = null;
        if (startingBlock < endingBlock)
          partialList = new long[] { startingBlock, endingBlock };
        else
          partialList = new long[] { startingBlock };
        for (int i = 0; i < partialList.length; i++)
        {
          long curBlockNumber = partialList[i];

          // Read the block status
          int curStatus = getBlockStatus(clearLevel, curBlockNumber);
          
          //todo: need to propagate up and check if any parents are complete
          if (curStatus != STATUS_UNREAD)
          {
            int[] newStatuses = null;
            if (curStatus == STATUS_COMPLETE)
            {
              // The STATUS_COMPLETE status needs to be pushed down to the children since the parent will no longer be STATUS_COMPLETE
              newStatuses = new int[childrenPerNode];
              long childSectorNumberStart = calcSectorNumber(clearLevel, curBlockNumber);
              for (int j = 0; j < childrenPerNode; j++)
              {
                long childSectorNumberEnd = childSectorNumberStart + (numSectorsPerBlock >>> childrenPerNodeLog);
                // Children not affected by the clear should be marked complete
                // Children in the clear should be left alone (status -1)
                if ((childSectorNumberEnd <= clearSectorStart) || (childSectorNumberStart >= clearSectorEnd))
                {
                  // No overlap with the range to clear
                  newStatuses[j] = STATUS_COMPLETE;
                }
                else
                  newStatuses[j] = -1;
              }
            }
            
            // The status should be updated to partial or empty, depending on the children.
            // Since updateBlockStatus does something like this, we'll use it to query the children
            // Pick the first child and query what the parent status should be. We are guaranteed to not
            // be on the last level at this point because the sector level can't have partial blocks.
            // If some children need to be updated to STATUS_COMPLETE, that will be done here via newStatuses
            int newStatus = updateBlockStatus(clearLevel + 1, curBlockNumber >>> childrenPerNodeLog, -1, newStatuses);
            // If the status is different, we actually have to update
            if (curStatus != newStatus)
            {
              // Now actually update the block. We don't normally need to cascade updates since we'll be checking the next level.
              updateBlockStatus(clearLevel, curBlockNumber, newStatus, null);
            }
          }
          if (lastUpdate)
            return;
        }
      }
    }
    return;
  }
*/
  // Clears a range of map entries
  private void clearMapBlockRange(int clearLevel, long startBlock, long blockCount) throws IOException
  {
    long startingBlockNumber = mapFileLevelOffsets[clearLevel] + startBlock;
    long endingBlockNumber = startingBlockNumber + blockCount;
    long clearBufferSize = (blockCount + childrenPerNode - 1) >>> 2;
    if (clearBufferSize > 1048576)
      clearBufferSize = 1048576;
    byte[] clearBuffer = new byte[(int)clearBufferSize];
    while (startingBlockNumber < endingBlockNumber)
    {
      int startingShift = (int)(startingBlockNumber & 3);
      if (startingShift > 0)
      {
        int blocksToBlank = 4 - startingShift;
        startingShift <<= 1;
        mapFileSeek(startingBlockNumber >>> 2);
        int mapByte = mapFile.read();
        if ((startingBlockNumber + blocksToBlank) > endingBlockNumber)
        {
          blocksToBlank = (int)(endingBlockNumber - startingBlockNumber);
          mapByte &= ~(((1 << (blocksToBlank << 1)) - 1) << startingShift);
        }
        else
          mapByte &= ((1 << startingShift) - 1);
        mapFileSeek(startingBlockNumber >>> 2);
        mapFile.write(mapByte);
        startingBlockNumber += blocksToBlank;
        continue;
      }
      long blocksRemaining = (endingBlockNumber - startingBlockNumber);
      long bytesRemaining = blocksRemaining >>> 2;
      if (bytesRemaining == 0)
      {
        int remainderShift = ((int)blocksRemaining & 3) << 1;
        mapFileSeek(startingBlockNumber >>> 2);
        int mapByte = mapFile.read();
        mapByte &= ~((1 << remainderShift) - 1);
        mapFileSeek(startingBlockNumber >>> 2);
        mapFile.write(mapByte);
        startingBlockNumber = endingBlockNumber;
        continue;
      }
      mapFileSeek(startingBlockNumber >>> 2);
      int bytesToWrite = clearBuffer.length;
      if (bytesToWrite > bytesRemaining)
        bytesToWrite = (int)bytesRemaining;
      mapFile.write(clearBuffer, 0, bytesToWrite);
      startingBlockNumber += (bytesToWrite << 2);
    }
    return;
  }
  
  // Opens the input and computes the tree size.
  private void openInput() throws Exception
  {
    if (inputDevice == null)
    {
      inputDevice = getNewBlockDevice(!inputIsDevice);
      if (inputOverrideSectorSize > 0)
        inputDevice.setSectorSize(inputOverrideSectorSize);
      else
        if (inputDevice.getSectorSize() <= 1)
          inputDevice.setSectorSize(512);
      inputDevice.open(inputFilename, false, false, false, inputIsDevice || (inputOverrideSectorSize > 0));
    }

    if ((badBlockEmulatorList != null) && (badBlockEmulatorList.size() > 0))
    {
      BadBlockEmulator bbe = new BadBlockEmulator(inputDevice);
      bbe.loadList(badBlockEmulatorList);
      inputDevice = bbe;
    }
    if (inputOverrideSize > 0L)
      inputSize = inputOverrideSize;
    else
      inputSize = inputDevice.length();
    // If an input offset is specified, it only makes sense to change the size of the input if the input size was not specified
    if ((inputOffset > 0L) && (inputOverrideSize <= 0L))
    {
      if (inputSize < inputOffset)
        throw new ParamException("Offset into the input at which to start reading (" + inputOffset + ") is greater than the length of the input (" + inputSize + ")");
      inputSize -= inputOffset;
    }
    inputSectorSizeLog = inputDevice.getSectorSizeLog();
    inputSizeSectors = inputSize >>> inputSectorSizeLog;
    // I don't think it's necessary to warn if the input length does not exactly match a sector boundary (the excess bytes will be dropped)

    numLeavesLog = roundUpAndLog(inputSizeSectors);
    maxLevelNumber = numLeavesLog / childrenPerNodeLog;
    numLeaves = 1 << numLeavesLog;
    
    int _bufferSizeSectorsLog = checkPower2(bufferSize) - inputSectorSizeLog;
    if (_bufferSizeSectorsLog < 0)
      throw new ParamException("Buffer size (" + bufferSize + ") must be the input sector size (" + (1 << inputSectorSizeLog) + ") multiplied by a power of 2");
    bufferSizeSectorsLog = _bufferSizeSectorsLog;
    
    if ((initialBlockSize > 0L) && (endBlockSize > 0L) && (initialBlockSize < endBlockSize))
      throw new ParamException("The ending block size specified (" + endBlockSize + ") is greater than the initial block size specified (" + initialBlockSize + ")");

    if (initialBlockSize > 0L)
      startLevel = checkRequestedBlockSize(initialBlockSize, "Initial block size");
    else
      startLevel = 0;

    if (endBlockSize > 0L)
      endLevel = checkRequestedBlockSize(endBlockSize, "End block size");
    else
      endLevel = maxLevelNumber;

    return;
  }
  
  // Checks whether the requested block size is valid.
  private int checkRequestedBlockSize(long blockSize, String paramName) throws Exception
  {
    int blockSizeLog = checkPower2(blockSize);
    if (blockSizeLog < 0)
      throw new ParamException(paramName + " specified is not a power of 2: " + blockSize);
    blockSizeLog -= inputSectorSizeLog;
    if (blockSizeLog < 0)
      throw new ParamException(paramName + " specified (" + blockSize + ") can not be smaller than the input sector size (" + (1 << inputSectorSizeLog) + ")");
    if ((blockSizeLog % childrenPerNodeLog) != 0)
      throw new ParamException(paramName + " specified (" + blockSize + ") must be a equal to the input sector size (" + (1 << inputSectorSizeLog) + ") multiplied by a power of the children per node (" + childrenPerNode + ")");
    int newLevel = maxLevelNumber - (blockSizeLog / childrenPerNodeLog);
    if (newLevel < 0)
      throw new ParamException(paramName + " specified (" + blockSize + ") is too large; maximum is " + (1L << (inputSectorSizeLog + numLeavesLog)));
    return newLevel;
  }
  
  // Opens the output and preallocates it if needed
  private void openOutput() throws Exception
  {
    if (outputDevice == null)
    {
      outputDevice = getNewBlockDevice(!outputIsDevice);
      outputDevice.open(outputFilename, true, true, false, outputIsDevice);
    }

    long targetSize = outputOffset + inputSize;
    if (preallocateOutput && (outputDevice.length() < targetSize))
    {
      byte[] buffer = new byte[4194304];
      long outputLen = outputDevice.length();
      if (writeToSystemOut)
        System.out.println("Preallocating output, writing " + (targetSize - outputLen) + " bytes...");
      outputDevice.seek(outputLen);
      while ((outputLen = outputDevice.length()) < targetSize)
      {
        int writeAmount = buffer.length;
        if (writeAmount > (targetSize - outputLen))
          writeAmount = (int)(targetSize - outputLen);
        outputDevice.write(buffer, 0, writeAmount);
      }
      outputDevice.seek(outputOffset);
    }
    return;
  }
  
  /**
   * Returns the number of blocks in the specified level.
   */
  public long calcNumBlocksForLevel(int levelNumber)
  {
    return 1L << (levelNumber * childrenPerNodeLog);
  }
  
  /**
   * Returns the block size in sectors for the specified level.
   */
  public long calcBlockSizeInSectors(int levelNumber)
  {
    return calcSectorNumber(levelNumber, 1);
  }
  
  /**
   * Returns the sector number corresponding to the start of the specified block at the specified level.
   */
  public long calcSectorNumber(int levelNumber, long blockNumber)
  {
    return blockNumber << ((maxLevelNumber - levelNumber) * childrenPerNodeLog);
  }
  
  /**
   * Returns the block number that contains the specified sector at the specified level.
   */
  public long calcBlockNumber(int levelNumber, long sectorNumber)
  {
    return sectorNumber >>> ((maxLevelNumber - levelNumber) * childrenPerNodeLog);
  }
  
  // Reads bytes from the map file, filling in 0's for bytes requested past the end of the map (due to caching).
  private void readMapFileBytes(long byteOffset, byte[] b, int offset, int len) throws IOException
  {
    int fillLen = 0;
    if ((byteOffset + len) > mapFileDataBytes)
    {
      if (byteOffset >= mapFileDataBytes)
      {
        fillLen = len;
        len = 0;
      }
      else
      {
        fillLen = (int)(byteOffset + len - mapFileDataBytes);
        len -= fillLen;
      }
    }
    if (len > 0)
    {
      mapFileSeek(byteOffset);
      while (len > 0)
      {
        int bytesRead = mapFile.read(b, offset, len);
        if (bytesRead <= 0)
          throw new IOException("Unable to read from map file at offset " + byteOffset);
        byteOffset += bytesRead;
        offset += bytesRead;
        len -= bytesRead;
      }
    }
    for (int i = 0; i < fillLen; i++)
      b[offset++] = 0;
    return;
  }
  
  /**
   * Returns an uninitialized {@link org.blinkenbyte.io.BlockDevice BlockDevice} that may be architecture/OS dependent.
   */
  public static BlockDevice getNewBlockDevice(boolean forceNormal) throws Exception
  {
    BlockDevice bd = null;
    String osName = System.getProperty("os.name");
    String osArch = System.getProperty("os.arch");
    if (!forceNormal && osName.toLowerCase().startsWith("windows") && osArch.toLowerCase().startsWith("x86"))
      bd = (BlockDevice)Class.forName("org.blinkenbyte.io.WindowsBlockDevice").newInstance();
    else
      bd = new NormalBlockDevice();
    return bd;
  }
  
  // Returns the exponent of the smallest power of childrenPerNode that is greater or equal to n.
  private int roundUpAndLog(long n)
  {
    if (n == 0)
      return -1;
    int bitPos = 0;
    long tempLong = n;
    while (tempLong != 1)
    {
      tempLong >>>= 1;
      bitPos++;
    }
    long highestOneBit = 1 << bitPos;
    if ((bitPos % childrenPerNodeLog) == 0)
    {
      if (n == highestOneBit)
        return bitPos;
      return bitPos + childrenPerNodeLog;
    }
    return bitPos + childrenPerNodeLog - (bitPos % childrenPerNodeLog);
  }
  
  /**
   * Checks whether the a number is a power of 2
   *
   * @param n The number to be checked.
   * @return Returns -1 if the number is not a power of 2; otherwise, returns the base 2 logarithm of the number.
   */
  public static int checkPower2(long n)
  {
    if (n <= 0)
      return -1;
    int logValue = 0;
    while ((n & 1) == 0)
    {
      n >>>= 1;
      logValue++;
    }
    if (n != 1)
      return -1;
    return logValue;
  }
  
  /**
   * Parses command-line arguments. Some parameters are set in this object, others are returned in the ParamInfo object
   */
  public ParamInfo parseArgs(final String[] argv)
  {
    ParamInfo paramInfo = new ParamInfo();
    Iterator<String> paramIterator = new Iterator<String>() {
      int curIndex = 0;
      
      public boolean hasNext()
      {  return (curIndex < argv.length);  }
      
      public String next()
      {  return hasNext() ? argv[curIndex++] : null;  }
      
      public void remove()
      {  }
    };
    try
    {
      while (paramIterator.hasNext())
      {
        String paramName = paramIterator.next();
        if (paramName.equals("-i"))
        {
          inputFilename = paramIterator.next();
          if (inputFilename == null)
          {
            paramInfo.errorMsg = "Missing filename after -i";
            return paramInfo;
          }
          continue;
        }
        if (paramName.equals("-o"))
        {
          outputFilename = paramIterator.next();
          if (outputFilename == null)
          {
            paramInfo.errorMsg = "Missing filename after -o";
            return paramInfo;
          }
          continue;
        }
        if (paramName.equals("-m"))
        {
          mapFilename = paramIterator.next();
          if (mapFilename == null)
          {
            paramInfo.errorMsg = "Missing filename after -m";
            return paramInfo ;
          }
          continue;
        }
        if (paramName.equals("--map-memory"))
        {
          mapToMemory = true;
          continue;
        }
        if (paramName.equals("--input-sector-size"))
        {
          inputOverrideSectorSize = (int)parseNumber("inputSectorSize", paramIterator.next());
          continue;
        }
        if (paramName.equals("--input-offset"))
        {
          setInputOffset(parseNumber("inputOffset", paramIterator.next()));
          continue;
        }
        if (paramName.equals("--input-size"))
        {
          inputOverrideSize = parseNumber("inputSize", paramIterator.next());
          continue;
        }
        if (paramName.equals("--output-offset"))
        {
          setOutputOffset(parseNumber("outputOffset", paramIterator.next()));
          continue;
        }
        if (paramName.equals("--children-per-node"))
        {
          setChildrenPerNode((int)parseNumber("childrenPerNode", paramIterator.next()));
          continue;
        }
        if (paramName.equals("--initial-block-size"))
        {
          initialBlockSize = parseNumber("initialBlockSize", paramIterator.next());
          continue;
        }
        if (paramName.equals("--end-block-size"))
        {
          endBlockSize = parseNumber("endBlockSize", paramIterator.next());
          continue;
        }
        if (paramName.equals("--input-bad-sectors"))
        {
          paramInfo.inputBadSectorsFilename = paramIterator.next();
          if (paramInfo.inputBadSectorsFilename == null)
          {
            paramInfo.errorMsg = "Missing filename after --input-bad-sectors";
            return paramInfo ;
          }
          continue;
        }
        if (paramName.equals("--buffer-size"))
        {
          setBufferSize((int)parseNumber("bufferSize", paramIterator.next()));
          continue;
        }
        if (paramName.equals("--preallocate-output"))
        {
          preallocateOutput = true;
          continue;
        }
        if (paramName.equals("--display-interval"))
        {
          callbackInterval = parseNumber("displayIntervalSeconds", paramIterator.next()) * 1000L;
          continue;
        }
        if (paramName.equals("--output-incomplete-sectors"))
        {
          paramInfo.outputIncompleteSectorsFilename = paramIterator.next();
          if (paramInfo.outputIncompleteSectorsFilename == null)
          {
            paramInfo.errorMsg = "Missing filename after --output-incomplete-sectors";
            return paramInfo;
          }
          continue;
        }
        if (paramName.equals("--force-input-device"))
        {
          if ((inputIsDeviceOverride != null) && !inputIsDeviceOverride.booleanValue())
          {
            paramInfo.errorMsg = "Conflicting options specified for forcing input to be a device";
            return paramInfo;
          }
          inputIsDeviceOverride = Boolean.TRUE;
          continue;
        }
        if (paramName.equals("--force-input-not-device"))
        {
          if ((inputIsDeviceOverride != null) && inputIsDeviceOverride.booleanValue())
          {
            paramInfo.errorMsg = "Conflicting options specified for forcing input to be a device";
            return paramInfo;
          }
          inputIsDeviceOverride = Boolean.FALSE;
          continue;
        }
        if (paramName.equals("--force-output-device"))
        {
          if ((outputIsDeviceOverride != null) && !outputIsDeviceOverride.booleanValue())
          {
            paramInfo.errorMsg = "Conflicting options specified for forcing output to be a device";
            return paramInfo;
          }
          outputIsDeviceOverride = Boolean.TRUE;
          continue;
        }
        if (paramName.equals("--force-output-not-device"))
        {
          if ((outputIsDeviceOverride != null) && outputIsDeviceOverride.booleanValue())
          {
            paramInfo.errorMsg = "Conflicting options specified for forcing output to be a device";
            return paramInfo;
          }
          outputIsDeviceOverride = Boolean.FALSE;
          continue;
        }
        if (paramName.equals("--clear-starting-level"))
        {
          paramInfo.mapClearStartingLevel = (int)parseNumber("clearStartingLevel", paramIterator.next());
          continue;
        }
        if (paramName.equals("--clear-block-range"))
        {
          paramInfo.mapClearBlockStart = parseNumber("clearStartBlock", paramIterator.next());
          paramInfo.mapClearBlockCount = parseNumber("clearBlockCount", paramIterator.next());
          continue;
        }
        if (paramName.equals("--clear-sector-range"))
        {
          paramInfo.clearSectorStart = parseNumber("clearSectorStart", paramIterator.next());
          paramInfo.clearSectorCount = parseNumber("clearSectorCount", paramIterator.next());
          continue;
        }
        if (paramName.equals("--read-map"))
        {
          if (paramInfo.clearMapBlocks || paramInfo.clearMapSectors)
          {
            paramInfo.errorMsg = "Read and clear map can not be requested at the same time";
            return paramInfo;
          }
          paramInfo.readMap = true;
          continue;
        }
        if (paramName.equals("--clear-map-blocks"))
        {
          if (paramInfo.readMap)
          {
            paramInfo.errorMsg = "Read and clear map can not be requested at the same time";
            return paramInfo;
          }
          if (paramInfo.clearMapSectors)
          {
            paramInfo.errorMsg = "Clear map by blocks and sectors can not be requested at the same time";
            return paramInfo;
          }
          paramInfo.clearMapBlocks = true;
          continue;
        }
        if (paramName.equals("--clear-map-sectors"))
        {
          if (paramInfo.readMap)
          {
            paramInfo.errorMsg = "Read and clear map can not be requested at the same time";
            return paramInfo;
          }
          if (paramInfo.clearMapBlocks)
          {
            paramInfo.errorMsg = "Clear map by blocks and sectors can not be requested at the same time";
            return paramInfo;
          }
          paramInfo.clearMapSectors = true;
          continue;
        }
        paramInfo.errorMsg = "Unrecognized parameter passed: " + paramName;
        return paramInfo;
      }
      if (paramInfo.readMap || paramInfo.clearMapBlocks || paramInfo.clearMapSectors)
      {
        if (mapFilename == null)
        {
          paramInfo.errorMsg = "No map filename given for --read-map";
          return paramInfo;
        }
        paramInfo.parsedOkay = true;
        return paramInfo;
      }
      if (inputFilename == null)
      {
        paramInfo.errorMsg = "No input filename given";
        return paramInfo;
      }
      if (inputIsDeviceOverride == null)
      {
        if (inputFilename.length() > 4)
          inputIsDevice = inputFilename.startsWith("\\\\?\\") || inputFilename.startsWith("/dev/");
      }
      else
        inputIsDevice = inputIsDeviceOverride.booleanValue();
      if (outputIsDeviceOverride != null)
        outputIsDevice = outputIsDeviceOverride.booleanValue();
      if (outputFilename == null)
      {
        paramInfo.errorMsg = "No output filename given";
        return paramInfo;
      }
      if (mapToMemory)
      {
        if (mapFilename != null)
        {
          paramInfo.errorMsg = "Memory-resident map requested, but a map filename also given";
          return paramInfo;
        }
      }
      else
      {
        if (mapFilename == null)
        {
          paramInfo.errorMsg = "No map filename given";
          return paramInfo;
        }
      }
      paramInfo.parsedOkay = true;
      return paramInfo;
    }
    catch (NumberFormatException nfe)
    {
      paramInfo.errorMsg = nfe.getMessage();
      return paramInfo;
    }
    catch (ParamException pe)
    {
      paramInfo.errorMsg = pe.getMessage();
      return paramInfo;
    }
  }
  
  private static long parseNumber(String paramName, String s) throws NumberFormatException
  {
    if ((s == null) || (s.length() == 0))
      throw new NumberFormatException("Blank number for parameter: " + paramName);
    try
    {
      return parseNumber(s);
    }
    catch (NumberFormatException nfe)
    {
      throw new NumberFormatException("Invalid or missing number for parameter: " + paramName);
    }
  }
  
  private static long parseNumber(String s) throws NumberFormatException
  {
    char c = s.charAt(s.length() - 1);
    int shifter = 0;
    switch (c)
    {
    case 'k':
    case 'K':
      shifter = 10;
      break;
    
    case 'm':
    case 'M':
      shifter = 20;
      break;
    
    case 'g':
    case 'G':
      shifter = 30;
      break;
    }
    if (shifter > 0)
      s = s.substring(0, s.length() - 1);
    long n = Long.parseLong(s);
    if (shifter > 0)
      n <<= shifter;
    return n;
  }
  
  /**
   * Command-line interface for the program.
   */
  public static void main(String[] argv)
  {
    if (argv.length == 0)
    {
      showInstructions();
      return;
    }
    ProgressiveCopier pc = new ProgressiveCopier();
    ParamInfo paramInfo = pc.parseArgs(argv);
    if (!paramInfo.parsedOkay)
    {
      System.out.println(paramInfo.errorMsg);
      return;
    }
    pc.writeToSystemOut = true;
    if (pc.callbackInterval >= 0)
      pc.callback = new DisplayStatus(pc);
doIt:
    try
    {
      if (paramInfo.readMap)
      {
        pc.openMapOnly();
        pc.writeIncompleteSectorsList(paramInfo.outputIncompleteSectorsFilename);
        System.out.println("maxLevelNumber=" + pc.getMaxLevelNumber() + ", childrenPerNode=" + pc.getNumChildrenPerNode());
        System.out.println("Done");
        break doIt;
      }
      if (paramInfo.clearMapBlocks)
      {
        pc.openMapOnly();
        System.out.println("Clearing " + ((paramInfo.mapClearBlockCount < 0) ? "all" : paramInfo.mapClearBlockCount) +
                           " blocks starting from block " + paramInfo.mapClearBlockStart + " at level " + paramInfo.mapClearStartingLevel);
        pc.clearMapBlocks(paramInfo.mapClearStartingLevel, paramInfo.mapClearBlockStart, paramInfo.mapClearBlockCount);
        System.out.println("Done");
        break doIt;
      }
      if (paramInfo.clearMapSectors)
      {
        pc.openMapOnly();
        if (paramInfo.clearSectorStart < 0)
        {
          paramInfo.clearSectorStart = 0;
          paramInfo.clearSectorCount = pc.getNumLeaves();
        }
        System.out.println("Clearing " + paramInfo.clearSectorCount + " sectors from map starting from sector " + paramInfo.clearSectorStart);
        pc.clearMapSectors(paramInfo.clearSectorStart, paramInfo.clearSectorCount);
        System.out.println("Done");
        break doIt;
      }

      if (paramInfo.inputBadSectorsFilename != null)
      {
        FileReader fr = null;
        try
        {
          fr = new FileReader(paramInfo.inputBadSectorsFilename);
          pc.emulateBadSectors(fr);
        }
        finally
        {
          if (fr != null)
          {
            try
            {  fr.close();  }
            catch (Exception e)
            {  }
            fr = null;
          }
        }
      
      }
      pc.open();
      pc.performCopy();
      if (paramInfo.outputIncompleteSectorsFilename != null)
        pc.writeIncompleteSectorsList(paramInfo.outputIncompleteSectorsFilename);
      System.out.println("Done");
    }
    catch (ParamException pe)
    {
      System.out.println(pe.getMessage());
    }
    catch (Exception e)
    {
      e.printStackTrace(System.out);
    }
    finally
    {
      pc.close();
    }
    return;
  }
  
  private static void showInstructions()
  {
    System.out.println("ProgressiveCopier -i inputFile -o outputFile {-m mapFile | --map-memory}");
    System.out.println("    [--input-sector-size inputSectorSize] [--input-offset inputOffset]");
    System.out.println("    [--input-size inputSize] [--output-offset outputOffset]");
    System.out.println("    [--children-per-node childrenPerNode]");
    System.out.println("    [--initial-block-size initialBlockSize] [--end-block-size endBlockSize]");
    System.out.println("    [--input-bad-sectors inputBadSectorsFile] [--buffer-size bufferSize]");
    System.out.println("    [--preallocate-output] [--display-interval displayIntervalSeconds]");
    System.out.println("    [--output-incomplete-sectors incompleteSectorsFilename]");
    System.out.println("    [--force-input-device | --force-input-not-device]");
    System.out.println("    [--force-output-device | --force-output-not-device]");
    System.out.println("ProgressiveCopier --read-map -m mapFile");
    System.out.println("    [--output-incomplete-sectors incompleteSectorsFilename]");
    System.out.println("ProgressiveCopier --clear-map-blocks -m mapFile");
    System.out.println("    [--clear-starting-level clearStartingLevel]");
    System.out.println("    [--clear-block-range clearStartBlock clearBlockCount]");
    System.out.println("ProgressiveCopier --clear-map-sectors -m mapFile");
    System.out.println("    [--clear-sector-range clearSectorStart clearSectorCount]");
    System.out.println();
    System.out.println("inputFile: The input file or device");
    System.out.println("outputFile: The output file or device");
    System.out.println("mapFile: The map file name");
    System.out.println("--map-memory: Create the map file in memory");
    System.out.println("inputSectorSize: Override input device's sector size. Must be power of 2");
    System.out.println("inputOffset: Offset into the input device to start reading (default 0)");
    System.out.println("inputSize: Override size of the input device; size relative to offset");
    System.out.println("outputOffset: Offset into the output device to start writing (default 0)");
    System.out.println("childrenPerNode: Number of pieces to divide blocks into. Default 4");
    System.out.println("initialBlockSize: The initial block size to use in the copy");
    System.out.println("endBlockSize: The smallest block size to use in the copy");
    System.out.println("inputBadSectorsFile: Optional input file of bad sectors in the input");
    System.out.println("  Each line contains the offset in sectors followed by the length in sectors");
    System.out.println("bufferSize: Input buffer size, power of two multiple of the sector size");
    System.out.println("preallocate-output: Create the output file's full size before starting");
    System.out.println("displayIntervalSeconds: Minimum interval at which to display progress");
    System.out.println("incompleteSectorsFilename: Name of file to save list of incomplete sectors to");
    System.out.println("--force-input[-not]-device: Force whether input is handled as a device");
    System.out.println("--force-output[-not]-device: Force whether output is handled as a device");
    System.out.println("--read-map: Only display information about the map file");
    System.out.println("--clear-map-blocks: Clear a range of blocks and sub-blocks from the map");
    System.out.println("clearStartingLevel: Starting level to clear data from map file");
    System.out.println("clearStartBlock: Starting block number to clear");
    System.out.println("clearBlockCount: Number of blocks to clear");
    System.out.println("--clear-map-sectors: Clear a range of sectors from the map");
    System.out.println("clearSectorStart: Starting sector number to clear");
    System.out.println("clearSectorCount: Number of sectors to clear");
    return;
  }
  
  /**
   * This class is used by the command-line program as a callback which displays basic information about the copy progress.
   */
  public static class DisplayStatus implements Runnable
  {
    public ProgressiveCopier pc = null;
    
    public DisplayStatus(ProgressiveCopier _pc)
    {
      pc = _pc;
    }
    
    public void run()
    {
      int curLevel = pc.getCurLevel();
      long blockSizeInSectors = pc.calcBlockSizeInSectors(curLevel);
      System.out.printf("Level %d/%d, block %d/%d, block size=%,d sectors (%,d bytes)\n", curLevel, pc.getMaxLevelNumber(), pc.getCurBlock(), pc.calcNumBlocksForLevel(curLevel), blockSizeInSectors, blockSizeInSectors << pc.getInputSectorSizeLog());
      return;
    }
  }
  
  /**
   * An exception class for bad parameters
   */
  public static class ParamException extends Exception
  {
    public ParamException(String msg)
    {
      super(msg);
    }
  }
  
  /**
   * Miscellaneous class for holding some command line parameters and parsing status
   */
  public static class ParamInfo
  {
    /**
     * True if the parsing operation succeeded, false otherwise.
     */
    public boolean parsedOkay = false;
    
    /**
     * The parsing error message if parsedOkay is false.
     */
    public String errorMsg = null;
    
    /**
     * The filename of the input bad sectors list.
     */
    public String inputBadSectorsFilename = null;
    
    /**
     * The filename of the output incomplete sectors list.
     */
    public String outputIncompleteSectorsFilename = null;
    
    /**
     * Set to true if the map only needs to be read.
     */
    public boolean readMap = false;
    
    /**
     * Set to true if a clear of the map by blocks is requested.
     */
    public boolean clearMapBlocks = false;
    
    /**
     * Set to true if a clear of the map by sectors is requested.
     */
    public boolean clearMapSectors = false;

    /**
     * The starting level for a map clear.
     */
    public int mapClearStartingLevel = 0;
    
    /**
     * The starting block for a map clear.
     */
    public long mapClearBlockStart = 0;
    
    /**
     * The block count for a map clear.
     */
    public long mapClearBlockCount = -1;
    
    /**
     * The starting sector for a map clear.
     */
    public long clearSectorStart = -1;

    /**
     * The sector count for a map clear.
     */
    public long clearSectorCount = -1;
    
    public ParamInfo()
    {  }
  }
  
}
